#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import os
import threading

from . import consts
from . import logging as pc_logging
from .mating import Mating
from . import project_config
from . import runtime_python_all
from . import project_factory_local as rfl
from . import project_factory_git as rfg
from . import project_factory_tar as rft
from . import sync_threads
from .user_config import user_config
from .utils import *
from .provider_request_quote import ProviderRequestQuote
from .provider_data_cart import *


# Context
class Context(project_config.Configuration):
    """Stores and caches all imported objects."""

    stats_packages: int
    stats_packages_instantiated: int
    stats_sketches: int
    stats_sketches_instantiated: int
    stats_interfaces: int
    stats_interfaces_instantiated: int
    stats_parts: int
    stats_parts_instantiated: int
    stats_assemblies: int
    stats_assemblies_instantiated: int
    stats_providers: int
    stats_provider_queries: int
    stats_memory: int

    # name is the package path (not a filesystem path) of the root package
    # in case it's configured to be something other than '/' (default)
    name: str

    # root_path is the absolute filesystem path to the root package (for monorepo's)
    root_path: str

    # current_project_path is the package path (not a filesystem path) of the current package
    # It is expected to match 'self.name' of the current package's object
    current_project_path: str

    class PackageLock(object):
        def __init__(self, ctx, package_name: str):
            ctx.project_locks_lock.acquire()
            if not package_name in ctx.project_locks:
                ctx.project_locks[package_name] = threading.Lock()
            self.lock = ctx.project_locks[package_name]
            ctx.project_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    def __init__(self, root_path=None, search_root=True):
        """Initializes the context and imports the root project."""
        root_file = ""
        if root_path is None:
            # Find the top folder containing "partcad.yaml"
            root_path = "."
        else:
            if os.path.isfile(root_path):
                root_file = os.path.basename(root_path)
                root_path = os.path.dirname(root_path)
        initial_root_path = os.path.abspath(root_path)
        if search_root:
            while os.path.exists(os.path.join(root_path, "..", "partcad.yaml")):
                root_path = os.path.join(root_path, "..")
        self.root_path = os.path.abspath(root_path)
        if self.root_path == initial_root_path and root_file != "":
            self.root_path = os.path.join(self.root_path, root_file)

        super().__init__(consts.ROOT, self.root_path)
        self.current_project_path = self.name
        if not self.current_project_path.endswith("/"):
            self.current_project_path += "/"
        self.current_project_path += os.path.relpath(
            initial_root_path,
            root_path,
        ).replace(os.path.sep, "/")
        if self.current_project_path == self.name + "/." or (
            self.name.endswith("/")
            and self.current_project_path == self.name + "."
        ):
            self.current_project_path = self.name

        # Protect the critical sections from access in different threads
        self.lock = threading.RLock()

        self.option_create_dirs = False
        self.runtimes_python = {}
        self.runtimes_python_lock = threading.Lock()

        self.stats_packages = 0
        self.stats_packages_instantiated = 0
        self.stats_interfaces = 0
        self.stats_interfaces_instantiated = 0
        self.stats_sketches = 0
        self.stats_sketches_instantiated = 0
        self.stats_parts = 0
        self.stats_parts_instantiated = 0
        self.stats_assemblies = 0
        self.stats_assemblies_instantiated = 0
        self.stats_providers = 0
        self.stats_provider_queries = 0
        self.stats_memory = 0

        self.mates = {}
        # self.projects contains all projects known to this context
        self.projects = {}
        self.project_locks = {}
        self.project_locks_lock = threading.Lock()
        self._projects_being_loaded = {}

        with pc_logging.Process("InitCtx", self.config_dir):
            self.import_project(
                None,  # parent
                {
                    "name": self.name,
                    "type": "local",
                    "path": self.config_path,
                    "maybeEmpty": True,
                    "isRoot": True,
                },
            )

    def stats_recalc(self, verbose=False):
        self.stats_memory = total_size(self, verbose)

    def get_current_project_path(self):
        return self.current_project_path

    def import_project(self, parent, project_import_config):
        if (
            not "name" in project_import_config
            or not "type" in project_import_config
        ):
            pc_logging.error(
                "Invalid project configuration found: %s"
                % project_import_config
            )
            return None

        name = project_import_config["name"]
        with self.PackageLock(self, name):
            with pc_logging.Action("Import", name):
                if name in self._projects_being_loaded:
                    pc_logging.error(
                        "Recursive project loading detected (%s), aborting."
                        % name
                    )
                    return None
                self._projects_being_loaded[name] = True

                # Depending on the project type, use different factories
                if (
                    not "type" in project_import_config
                    or project_import_config["type"] == "local"
                ):
                    with pc_logging.Action("Local", name):
                        rfl.ProjectFactoryLocal(
                            self, parent, project_import_config
                        )
                elif project_import_config["type"] == "git":
                    with pc_logging.Action("Git", name):
                        rfg.ProjectFactoryGit(
                            self, parent, project_import_config
                        )
                elif project_import_config["type"] == "tar":
                    with pc_logging.Action("Tar", name):
                        rft.ProjectFactoryTar(
                            self, parent, project_import_config
                        )
                else:
                    pc_logging.error("Invalid project type found: %s." % name)
                    del self._projects_being_loaded[name]
                    return None

                # Check whether the factory was able to successfully add the project
                if not name in self.projects:
                    pc_logging.error(
                        "Failed to create the project: %s"
                        % project_import_config
                    )
                    del self._projects_being_loaded[name]
                    return None

                imported_project = self.projects[name]
                if imported_project is None:
                    pc_logging.error("Failed to import the package: %s" % name)
                    del self._projects_being_loaded[name]
                    return None
                if imported_project.broken:
                    pc_logging.error(
                        "Failed to parse the package's 'partcad.yaml': %s"
                        % name
                    )

                self.stats_packages += 1
                self.stats_packages_instantiated += 1

                del self._projects_being_loaded[name]
                return imported_project

    def get_project_abs_path(self, rel_project_path: str):
        """
        Get the full package path (not a filesystem path) of a package
        given the relative path from the current package
        """
        if rel_project_path.startswith("/"):
            return rel_project_path

        project_path = self.current_project_path

        if rel_project_path == ".":
            rel_project_path = ""
        if rel_project_path == "":
            return project_path

        return get_child_project_path(project_path, rel_project_path)

    def get_project(self, rel_project_path: str):
        project_path = self.get_project_abs_path(rel_project_path)

        with self.lock:
            # Check if it's an explicit reference outside of the root project
            if not project_path.startswith(self.name):
                pc_logging.debug(
                    "Project path is outside of the root project: %s"
                    % project_path
                )
                # In case of an explicit reference outside of the root project,
                # assume that the root package has an 'onlyInRoot' dependency
                # present to facilitate such a reference in a standalone
                # development environment.

            # Strip the first '/' (absolute path always starts with a '/'``)
            if self.name != "/" and project_path.startswith(self.name):
                # The root package is not '/, need to skip the root package name
                len_to_skip = len(self.name) + 1
            else:
                len_to_skip = 1
            project_path = project_path[len_to_skip:]

            project = self.projects[self.name]

            if project_path == "":
                return project
            else:
                import_list = project_path.split("/")

            return self._get_project_recursive(project, import_list)

    def _get_project_recursive(self, project, import_list: list[str]):
        """Load the dependencies recursively"""
        if len(import_list) == 0:
            # Found what we are looking for
            return project

        # next_import is the next of the import we need to load now
        next_import = import_list[0]
        # import_list is reduced to contain only the items that will remain to
        # bt loaded after this import
        import_list = import_list[1:]

        # next_project will reference the project we are importing now
        next_project = None
        # next_project_path is the full path of the project we are importing now
        next_project_path = get_child_project_path(project.name, next_import)

        # see if the wanted project is already initialized
        if next_project_path in self.projects:
            return self._get_project_recursive(
                self.projects[next_project_path], import_list
            )

        # Check if there is a matching subfolder
        subfolders = [
            f.name for f in os.scandir(project.config_dir) if f.is_dir()
        ]
        if next_import in list(subfolders):
            if os.path.exists(
                os.path.join(
                    project.config_dir,
                    next_import,
                    consts.DEFAULT_PACKAGE_CONFIG,
                )
            ):
                pc_logging.debug(
                    "Importing a subfolder (get): %s..." % next_project_path
                )
                prj_conf = {
                    "name": next_project_path,
                    "type": "local",
                    "path": next_import,
                }
                next_project = self.import_project(project, prj_conf)
                if not next_project is None:
                    result = self._get_project_recursive(
                        next_project, import_list
                    )
                    return result
        else:
            # Otherwise, iterate all subfolders and check if any of them are packages
            if (
                "import" in project.config_obj
                and not project.config_obj["import"] is None
            ):
                imports = project.config_obj["import"]
                # TODO(clairbee): revisit if this code path is needed when the
                #                 user explicitly asked for a particular package
                # if not project.config_obj.get("isRoot", False):
                #     filtered = filter(
                #         lambda x: "onlyInRoot" not in imports[x]
                #         or not imports[x]["onlyInRoot"],
                #         imports,
                #     )
                #     imports = list(filtered)
                for prj_name in imports:
                    pc_logging.debug(
                        "Checking the import: %s vs %s..."
                        % (prj_name, next_import)
                    )
                    if prj_name != next_import:
                        continue
                    prj_conf = project.config_obj["import"][prj_name]
                    if prj_conf.get("onlyInRoot", False):
                        next_project_path = "/" + prj_name
                    pc_logging.debug("Importing: %s..." % next_project_path)
                    if "name" in prj_conf:
                        prj_conf["orig_name"] = prj_conf["name"]
                    prj_conf["name"] = next_project_path
                    next_project = self.import_project(project, prj_conf)
                    if not next_project is None:
                        result = self._get_project_recursive(
                            next_project, import_list
                        )
                        return result
                    break

        return next_project

    def import_all(self, parent_name=None):
        if parent_name is None:
            parent_name = self.name
        asyncio.run(self._import_all_wrapper(self.projects[parent_name]))

    async def _import_all_wrapper(self, project):
        iterate_tasks = []
        import_tasks = []

        iterate_tasks.append(
            asyncio.create_task(self._import_all_recursive(project))
        )

        while iterate_tasks or import_tasks:
            if iterate_tasks:
                iterate_done, iterate_tasks_set = await asyncio.tasks.wait(
                    iterate_tasks
                )
                iterate_tasks = list(iterate_tasks_set)
                for iterate_task in iterate_done:
                    new_import_tasks = iterate_task.result()
                    import_tasks.extend(new_import_tasks)

            if import_tasks:
                import_done, import_tasks_set = await asyncio.tasks.wait(
                    import_tasks
                )
                import_tasks = list(import_tasks_set)
                for import_task in import_done:
                    next_project = import_task.result()
                    iterate_tasks.append(
                        asyncio.create_task(
                            self._import_all_recursive(next_project)
                        )
                    )

    async def _import_all_recursive(self, project):
        tasks = []

        if project.broken:
            pc_logging.warn("Ignoring the broken package: %s" % project.name)
            return []

        # First, iterate all explicitly mentioned "import"s.
        # Do it before iterating subdirectories, as it may kick off a long
        # background task.
        if (
            "import" in project.config_obj
            and not project.config_obj["import"] is None
        ):
            imports = project.config_obj["import"]
            if not project.config_obj.get("isRoot", False):
                filtered = filter(
                    lambda x: "onlyInRoot" not in imports[x]
                    or not imports[x]["onlyInRoot"],
                    imports,
                )
                imports = list(filtered)

            for prj_name in imports:
                prj_conf = project.config_obj["import"][prj_name]

                if prj_conf.get("onlyInRoot", False):
                    next_project_path = "/" + prj_name
                else:
                    next_project_path = get_child_project_path(
                        project.name, prj_name
                    )
                if next_project_path == self.name:
                    # Avoid circular imports of the root package
                    # TODO(clairbee): fix circular imports in general
                    continue
                pc_logging.debug("Importing: %s..." % next_project_path)

                if "name" in prj_conf:
                    prj_conf["orig_name"] = prj_conf["name"]
                prj_conf["name"] = next_project_path

                tasks.append(
                    asyncio.create_task(
                        sync_threads.run(self.import_project, project, prj_conf)
                    )
                )

        # Second, iterate over all subfolder and check for packages
        subfolders = [
            f.name for f in os.scandir(project.config_dir) if f.is_dir()
        ]
        for subdir in list(subfolders):
            if os.path.exists(
                os.path.join(
                    project.config_dir,
                    subdir,
                    consts.DEFAULT_PACKAGE_CONFIG,
                )
            ):
                # TODO(clairbee): check if this subdir is already imported
                next_project_path = get_child_project_path(project.name, subdir)

                if next_project_path in self.projects:
                    # This project is already imported (e.g. the current one)
                    # TODO(clairbee): is this worth a warning?
                    pc_logging.debug("Skipping already imported: %s" % subdir)
                    continue

                pc_logging.debug(
                    "Importing a subfolder (import all): %s..."
                    % next_project_path
                )
                prj_conf = {
                    "name": next_project_path,
                    "type": "local",
                    "path": subdir,
                }

                tasks.append(
                    asyncio.create_task(
                        sync_threads.run(self.import_project, project, prj_conf)
                    )
                )

        return tasks

    def get_all_packages(self, parent_name=None):
        # TODO(clairbee): leverage root_project.get_child_project_names()
        self.import_all(parent_name)
        return self.get_packages(parent_name)

    def get_packages(self, parent_name=None):
        projects = self.projects.values()
        if parent_name is not None:
            projects = filter(
                lambda x: x.name.startswith(parent_name), projects
            )
        return map(
            lambda pkg: {"name": pkg.name, "desc": pkg.desc},
            filter(
                # FIXME(clairbee): parameterize interfaces before displaying them
                # lambda x: len(x.interfaces) + len(x.sketches) + len(x.parts) + len(x.assemblies)
                lambda x: len(x.sketches) + len(x.parts) + len(x.assemblies)
                > 0,
                projects,
            ),
        )

    def add_mate(
        self, source_interface, target_interface, mate_target_config: dict
    ):
        self._add_mate(
            source_interface,
            target_interface,
            mate_target_config,
            reverse=False,
        )
        if target_interface != source_interface:
            self._add_mate(
                target_interface,
                source_interface,
                mate_target_config,
                reverse=True,
            )

    def _add_mate(
        self,
        source_interface,
        target_interface,
        mate_target_config: dict,
        reverse: bool,
    ):
        source_interface_name = source_interface.full_name
        target_interface_name = target_interface.full_name

        if not source_interface_name in self.mates:
            self.mates[source_interface_name] = {}
        if target_interface_name in self.mates[source_interface_name]:
            pc_logging.error(
                "Mate already exists: %s -> %s"
                % (source_interface_name, target_interface_name)
            )
            return

        mate = Mating(
            source_interface, target_interface, mate_target_config, reverse
        )
        self.mates[source_interface_name][target_interface_name] = mate

    def get_mate(self, source_interface_name, target_interface_name):
        if not source_interface_name in self.mates:
            return None
        if not target_interface_name in self.mates[source_interface_name]:
            return None

        return self.mates[source_interface_name][target_interface_name]

    def find_mating_interfaces(self, source_shape, target_shape):
        source_interfaces = set(source_shape.with_ports.get_interfaces().keys())

        # compatible_source_interfaces is the list of all interfaces they are
        # compatible with (implement them and only them)
        compatible_source_interfaces = set(
            [
                compatible_interface
                for interface in source_interfaces
                for compatible_interface in self.get_interface(
                    interface
                ).compatible_with
            ]
        )

        # real_source_interfaces is the map of source interfaces to the set of
        # interfaces they are compatible with (including themselves). It's
        # called 'real', beacuase it allows to perform a lookup of
        # the actual (real) source interface(s) which brought in the given
        # compatible interface.
        real_source_interfaces = {
            interface: set([interface]) for interface in source_interfaces
        }
        for interface in source_interfaces:
            for compatible_interface in self.get_interface(
                interface
            ).compatible_with:
                if not compatible_interface in real_source_interfaces:
                    real_source_interfaces[compatible_interface] = set()
                real_source_interfaces[compatible_interface].add(interface)

        # Now, extend the source_interfaces to include all of the interfaces at
        # least one of the source interfaces is compatible with
        source_interfaces = source_interfaces.union(
            compatible_source_interfaces
        )
        pc_logging.debug("Source interfaces: %s" % source_interfaces)

        # Now do the same for the target interfaces
        target_interfaces = set(target_shape.with_ports.get_interfaces().keys())
        compatible_target_interfaces = set(
            [
                compatible_interface
                for interface in target_interfaces
                for compatible_interface in self.get_interface(
                    interface
                ).compatible_with
            ]
        )
        real_target_interfaces = {
            interface: set([interface]) for interface in target_interfaces
        }
        for interface in target_interfaces:
            for compatible_interface in self.get_interface(
                interface
            ).compatible_with:
                if not compatible_interface in real_target_interfaces:
                    real_target_interfaces[compatible_interface] = set()
                real_target_interfaces[compatible_interface].add(interface)
        target_interfaces = target_interfaces.union(
            compatible_target_interfaces
        )
        pc_logging.debug("Target interfaces: %s" % target_interfaces)

        source_interfaces_mates = set(
            [
                peer_interface
                for source_interface in source_interfaces
                for peer_interface in self.mates.get(
                    source_interface, {}
                ).keys()
            ]
        )
        target_interfaces_mates = set(
            [
                peer_interface
                for target_interface in target_interfaces
                for peer_interface in self.mates.get(
                    target_interface, {}
                ).keys()
            ]
        )
        source_candidate_interfaces = source_interfaces.intersection(
            target_interfaces_mates
        )
        real_source_candidate_interfaces = set()
        for source_interface in source_candidate_interfaces:
            real_source_candidate_interfaces = (
                real_source_candidate_interfaces.union(
                    real_source_interfaces[source_interface]
                )
            )
        source_candidate_interfaces = real_source_candidate_interfaces
        pc_logging.debug(
            "Source candidate interfaces: %s" % source_candidate_interfaces
        )

        target_candidate_interfaces = target_interfaces.intersection(
            source_interfaces_mates
        )
        real_target_candidate_interfaces = set()
        for target_interface in target_candidate_interfaces:
            real_target_candidate_interfaces = (
                real_target_candidate_interfaces.union(
                    real_target_interfaces[target_interface]
                )
            )
        target_candidate_interfaces = real_target_candidate_interfaces
        pc_logging.debug(
            "Target candidate interfaces: %s" % target_candidate_interfaces
        )

        return source_candidate_interfaces, target_candidate_interfaces

    def _get_sketch(self, sketch_spec, params=None):
        project_name, sketch_name = resolve_resource_path(
            self.current_project_path,
            sketch_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            pc_logging.error("Packages found: %s" % str(self.projects))
            return None
        pc_logging.debug("Retrieving %s from %s" % (sketch_name, project_name))
        return prj.get_sketch(sketch_name, params)

    def get_sketch(self, sketch_spec, params=None):
        return self._get_sketch(sketch_spec, params)

    def get_sketch_shape(self, sketch_spec, params=None):
        return asyncio.run(self._get_sketch(sketch_spec, params).get_wrapped())

    def get_sketch_cadquery(self, sketch_spec, params=None):
        return asyncio.run(self._get_sketch(sketch_spec, params).get_cadquery())

    def get_sketch_build123d(self, sketch_spec, params=None):
        return asyncio.run(
            self._get_sketch(sketch_spec, params).get_build123d()
        )

    def _get_interface(self, interface_spec):
        project_name, interface_name = resolve_resource_path(
            self.current_project_path,
            interface_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            pc_logging.error("Packages found: %s" % str(self.projects))
            return None
        pc_logging.debug(
            "Retrieving %s from %s" % (interface_name, project_name)
        )
        return prj.get_interface(interface_name)

    def get_interface(self, interface_spec):
        return self._get_interface(interface_spec)

    def get_interface_shape(self, interface_spec):
        return asyncio.run(self._get_interface(interface_spec).get_wrapped())

    async def find_suppliers(self, cart: ProviderCart) -> dict[str, list[str]]:
        """Find suppliers for each of the parts in the cart"""
        suppliers = {}
        for name, part_spec in cart.parts.items():
            if not ":" in name:
                name = ":" + name
            project_name, part_name = resolve_resource_path(
                self.current_project_path,
                name,
            )
            prj = self.get_project(project_name)
            if prj is None:
                pc_logging.error("Package %s not found" % project_name)
                pc_logging.error("Packages found: %s" % str(self.projects))
                return {}
            pc_logging.debug("Retrieving suppliers from %s" % project_name)

            part_suppliers = prj.get_suppliers(part_spec)
            if len(part_suppliers) == 0:
                pc_logging.error(
                    "No suppliers found for %s in %s"
                    % (part_name, project_name)
                )
                return {}

            pc_logging.debug("Part spec: %s" % str(part_spec))
            for provider_name, provider_extra_config in part_suppliers.items():
                provider = self.get_provider(
                    provider_name, provider_extra_config
                )
                if cart.qos is not None and not provider.is_qos_available(
                    cart.qos
                ):
                    continue
                if not await provider.is_part_available(part_spec):
                    continue

                if f"{project_name}:{part_name}" not in suppliers:
                    suppliers[f"{project_name}:{part_name}"] = []
                suppliers[f"{project_name}:{part_name}"].append(provider.name)

            if f"{project_name}:{part_name}" not in suppliers:
                pc_logging.error(
                    f"No supplier found for {project_name}:{part_name}"
                )

        # TODO(clairbee): calculate the recommended suppliers and
        #                 reorder the results accordingly

        return suppliers

    def select_preferred_supplier(self, suppliers: list[str]):
        """From a list of suppliers, select the preferred one."""
        return suppliers[0]

    def select_preferred_suppliers(
        self, suppliers_per_part: dict[str, list[str]]
    ) -> dict[str, str]:
        """Given a list of suppliers per part, select the preferred supplier for each."""
        preferred_suppliers = {}
        for name, suppliers in suppliers_per_part.items():
            supplier = self.select_preferred_supplier(suppliers)
            preferred_suppliers[name] = supplier

        return preferred_suppliers

    async def select_supplier(
        self, provider, cart: ProviderCart
    ) -> dict[str, str]:
        """Given a specific provider, confirm it can provide all the parts."""
        if cart.qos is not None and not provider.is_qos_available(cart.qos):
            pc_logging.error(f"QoS {cart.qos} is not available from {provider}")
            return
        suppliers = {}
        tasks = []

        async def _set_supplier(name, cart_item):
            if not await provider.is_part_available(cart_item):
                pc_logging.error(
                    "Part %s is not available from %s" % (name, provider.name)
                )
                suppliers[str(cart_item)] = ""
            else:
                pc_logging.debug("Cart item: %s" % str(cart_item))
                suppliers[str(cart_item)] = provider.name

        for name, cart_item in cart.parts.items():
            tasks.append(asyncio.create_task(_set_supplier(name, cart_item)))
        await asyncio.gather(*tasks)

        return suppliers

    async def prepare_supplier_carts(
        self, preferred_suppliers: dict[str, str]
    ) -> dict[str, ProviderCart]:
        """Given the list of preferred suppliers, prepare the supplier carts."""
        supplier_carts: dict[str, ProviderCart] = {}

        # Create a supplier cart for each supplier
        provider_names = set(preferred_suppliers.values())
        for provider_name in provider_names:
            supplier_carts[provider_name] = ProviderCart()

        # Place each part in the corresponding supplier cart
        for part_spec, provider_name in preferred_suppliers.items():
            if not provider_name in supplier_carts:
                supplier_carts[provider_name] = ProviderCart()
            cart_item = await supplier_carts[provider_name].add_part_spec(
                self, part_spec
            )

            # If it's not 'None' which means no provider found
            if provider_name:
                # Load the provider-supported CAD mode.
                # This is what makes it a 'supplier cart' instead of a regular cart.
                provider = self.get_provider(provider_name)
                await provider.load(cart_item)

        return supplier_carts

    async def supplier_carts_to_quotes(
        self, preferred_suppliers: dict[str, ProviderCart]
    ) -> dict[str, ProviderRequestQuote]:
        """Given the list of suppliers, and a supplier-customized carts,
        convert them to quotes"""
        quotes = {}
        for provider_name, supplier_cart in preferred_suppliers.items():

            pc_logging.debug("Supplier: %s" % str(provider_name))

            quote = ProviderRequestQuote(supplier_cart)

            if provider_name:
                # If it's not 'None' which means no provider found
                provider = self.get_provider(provider_name)
                quote_result = await provider.query_quote(quote)
                quote.set_result(quote_result)
            quotes[provider_name] = quote
        return quotes

    def get_provider(self, part_spec, params=None):
        project_name, part_name = resolve_resource_path(
            self.current_project_path,
            part_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            pc_logging.error("Packages found: %s" % str(self.projects))
            return None
        pc_logging.debug("Retrieving %s from %s" % (part_name, project_name))
        return prj.get_provider(part_name, params)

    def _get_part(self, part_spec, params=None):
        project_name, part_name = resolve_resource_path(
            self.current_project_path,
            part_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            pc_logging.error("Packages found: %s" % str(self.projects))
            return None
        pc_logging.debug("Retrieving %s from %s" % (part_name, project_name))
        return prj.get_part(part_name, params)

    def get_part(self, part_spec, params=None):
        return self._get_part(part_spec, params)

    def get_part_shape(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_wrapped())

    def get_part_cadquery(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_cadquery())

    def get_part_build123d(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_build123d())

    def _get_assembly(self, assembly_spec, params=None):
        project_name, assembly_name = resolve_resource_path(
            self.current_project_path,
            assembly_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            return None
        pc_logging.debug(
            "Retrieving %s from %s" % (assembly_name, project_name)
        )
        return prj.get_assembly(assembly_name, params)

    def get_assembly(self, assembly_spec, params=None):
        return self._get_assembly(assembly_spec, params)

    def get_assembly_shape(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_wrapped()
        )

    def get_assembly_cadquery(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_cadquery()
        )

    def get_assembly_build123d(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_build123d()
        )

    def render(self, project_path=None, format=None, output_dir=None):
        if project_path is None:
            project_path = self.get_current_project_path()
        pc_logging.info("Rendering all objects in %s..." % project_path)
        project = self.get_project(project_path)
        project.render(format=format, output_dir=output_dir)

    def get_python_runtime(self, version=None, python_runtime=None):
        with self.runtimes_python_lock:
            if version is None:
                version = "%d.%d" % (
                    sys.version_info.major,
                    sys.version_info.minor,
                )
            if python_runtime is None:
                python_runtime = user_config.python_runtime
            runtime_name = python_runtime + "-" + version
            if not runtime_name in self.runtimes_python:
                self.runtimes_python[runtime_name] = runtime_python_all.create(
                    self, version, python_runtime
                )
            return self.runtimes_python[runtime_name]

    def ensure_dirs(self, path):
        if not self.option_create_dirs:
            return
        os.makedirs(path)

    def ensure_dirs_for_file(self, filename):
        if not self.option_create_dirs:
            return
        path = os.path.dirname(filename)
        os.makedirs(path)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within the python sandbox environment (python runtime)
# to invoke `cadquery` scripts.

import os
import re
import sys

from cadquery import cqgi

sys.path.append(os.path.dirname(__file__))
import wrapper_common
import py_stubs.ocp_vscode  # Make 'sys.modules["py_stub.ocp_vscode"]' available


def process(path, request):
    build_parameters = {}
    if "patch" in request:
        patch = request["patch"]
    else:
        patch = {}
    if "build_parameters" in request:
        build_parameters = request["build_parameters"]

    script = open(path, "r").read()
    for old, new in patch.items():
        script = re.sub(old, new, script, flags=re.MULTILINE)

    if "import partcad" in script:
        script = (
            "import logging\nlogging.basicConfig(level=60)\n"  # Disable PartCAD logging
            + script
        )

    # Ignore ocp_vscode as it is of no use in the sandboxed environment
    # and it produces a lot of sporadic output to stdout and stderr
    sys.modules["ocp_vscode"] = sys.modules["py_stubs.ocp_vscode"]
    if "from ocp_vscode import " in script:
        script = re.sub(
            r"(from ocp_vscode import .*)\n",
            "saved_show_object_early=show_object\n\\1\nimport ocp_vscode\nocp_vscode.saved_show_object=saved_show_object_early\n",
            script,
        )
    if "import ocp_vscode" in script:
        script = script.replace(
            "import ocp_vscode",
            "import ocp_vscode\nocp_vscode.saved_show_object=show_object\n#",
        )

    # Execute the script
    script_object = cqgi.parse(script)
    build_result = script_object.build(build_parameters=build_parameters)

    if not build_result.success:
        wrapper_common.handle_exception(build_result.exception, path)

    shapes = []
    for result in build_result.results:
        shape = result.shape
        if hasattr(shape, "toOCC"):
            shape = shape.toOCC()
        if hasattr(shape, "val"):
            shape = shape.val()
        if hasattr(shape, "toCompound"):
            shape = shape.toCompound()
        if hasattr(shape, "wrapped"):
            shape = shape.wrapped
        shapes.append(shape)

    return {
        "success": build_result.success,
        "exception": build_result.exception,
        "shapes": shapes,
    }


path, request = wrapper_common.handle_input()

# Call CadQuery
model = process(path, request)

wrapper_common.handle_output(model)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import cadquery as cq

import base64
import os
import pickle
import sys
import threading
import time

from . import logging as pc_logging
from . import wrapper
from .part_factory_file import PartFactoryFile

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import register as register_cq_helper


class PartFactoryStep(PartFactoryFile):
    # How many STEP files should be loaded directly simultaneously (without
    # launching a subprocess), no matter the file size.
    MIN_SIMPLE_INFLIGHT = 1

    # How big of a STEP file is needed to consider launching a sub-process.
    MIN_SUBPROCESS_FILE_SIZE = 64 * 1024

    # lock is used to protect changes to the static class members
    lock = threading.Lock()
    # How many of the current instances are in which mode?
    count_inflight_simple = 0
    count_inflight_subprocess = 0

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action("InitSTEP", target_project.name, config["name"]):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                extension=".step",
            )
            # Complement the config object here if necessary
            self._create(config)

            self.runtime = None

    async def instantiate(self, part):
        await super().instantiate(part)

        with pc_logging.Action("STEP", part.project_name, part.name):
            do_subprocess = False

            file_size = os.path.getsize(self.path)

            with PartFactoryStep.lock:
                if (
                    PartFactoryStep.count_inflight_simple
                    < PartFactoryStep.MIN_SIMPLE_INFLIGHT
                    or file_size < PartFactoryStep.MIN_SUBPROCESS_FILE_SIZE
                ):
                    PartFactoryStep.count_inflight_simple += 1
                else:
                    do_subprocess = True
                    PartFactoryStep.count_inflight_subprocess += 1
                    if self.runtime == None:
                        # We don't care about customer preferences much here
                        # as this is expected to be hermetic.
                        # Stick to the version where CadQuery is known to work.
                        self.runtime = self.ctx.get_python_runtime("3.10")

            if do_subprocess:
                wrapper_path = wrapper.get("step.py")

                request = {"build_parameters": {}}
                register_cq_helper()
                picklestring = pickle.dumps(request)
                request_serialized = base64.b64encode(picklestring).decode()

                await self.runtime.ensure("cadquery")
                response_serialized, errors = await self.runtime.run(
                    [
                        wrapper_path,
                        os.path.abspath(self.path),
                        os.path.abspath(self.project.config_dir),
                    ],
                    request_serialized,
                )
                sys.stderr.write(errors)

                response = base64.b64decode(response_serialized)
                register_cq_helper()
                result = pickle.loads(response)

                if not result["success"]:
                    pc_logging.error(result["exception"])
                    raise Exception(result["exception"])
                shape = result["shape"]
            else:
                # Thanks for OCP deficiencies in handling of GIL,
                # as soon as 'importStep()' is called, all of other Python
                # threads are going to be frozen, so we need to give other
                # threads an opportunity to spawn processes to stay busy during
                # the 'importStep()' call.
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                time.sleep(0.0001)
                # TODO(clairbee): remove sleep calls when GIL is fixed in CQ

                shape = cq.importers.importStep(self.path).val().wrapped

            with PartFactoryStep.lock:
                if do_subprocess:
                    PartFactoryStep.count_inflight_subprocess -= 1
                else:
                    PartFactoryStep.count_inflight_simple -= 1

            self.ctx.stats_parts_instantiated += 1

            return shape



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-01
#
# Licensed under Apache License, Version 2.0.
#

# This script contains code shared by all wrapper scripts.

import base64

# import fcntl  # TODO(clairbee): replace it with whatever works on Windows if needed
import os
import pickle
import sys

from cq_serialize import register as register_cq_helper


def handle_input():
    if len(sys.argv) < 2:
        sys.stderr.write("Usage: %s <path>\n" % sys.argv[0])
        sys.exit(1)

    # Handle the input
    # - Comand line parameters
    path = os.path.normpath(sys.argv[1])
    if len(sys.argv) > 2:
        os.chdir(os.path.normpath(sys.argv[2]))
    # - Content passed via stdin
    # #   - Make stdin blocking so that we can read until EOF
    # flag = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    # fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flag & ~os.O_NONBLOCK)
    #   - Read until EOF
    input_str = sys.stdin.read()
    #   - Unpack the content received via stdin

    # TODO(clairbee): is .encode() needed here?
    request_bytes = base64.b64decode(input_str)

    register_cq_helper()
    request = pickle.loads(request_bytes)
    return path, request


def handle_output(model):
    # Serialize the output
    register_cq_helper()
    picklestring = pickle.dumps(model)
    response = base64.b64encode(picklestring)
    sys.stdout.write(response.decode())
    sys.stdout.flush()


def handle_exception(exc, cqscript=None):
    sys.stderr.write("Error: [")
    sys.stderr.write(str(exc).strip())
    sys.stderr.write("] on the line: [")
    tb = exc.__traceback__
    # Switch to the traceback object that contains the script line number
    tb = tb.tb_next
    # Get the filename
    fname = tb.tb_frame.f_code.co_filename
    if cqscript is not None and fname == "<cqscript>":
        fname = cqscript

    # Get the line contents
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
        if tb.tb_lineno - 1 < len(lines):
            line = lines[tb.tb_lineno - 1]
            sys.stderr.write(line.strip())
    sys.stderr.write("]\n")
    sys.stderr.flush()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import copy

from . import logging as pc_logging
from .utils import resolve_resource_path


def resolve_cart_item(item_spec: str):
    if "#" in item_spec:
        components = item_spec.split("#")
        name = components[0]
        assert len(components) == 2
        assert components[1].isdigit()
        count = int(components[1])
        assert count > 0
    else:
        # Without the count or parameters
        name = item_spec
        count = 1

    return name, count


class ProviderCartItem:
    """
    Describes a single part in a cart.
    It must be serializable for persistence.
    It must be deterministic: it must not be possible
    to imply a different CAD model after deserialization,
    even if the part definition has changed since it was serialized.
    """

    name: str
    count: int
    material: str = None
    color: str = None
    finish: str = None
    # TODO(clairbee): add texture
    # TODO(clairbee): add tolerance

    vendor: str = None
    sku: str = None
    count_per_sku: int = None

    format: str = None
    binary: bytes = None

    def __init__(self):
        self.name = "none"
        self.count = 0
        self.format = "none"

    async def set_spec(self, ctx, spec: str):
        self.name, self.count = resolve_cart_item(spec)

        part = ctx.get_part(self.name)
        assert part is not None
        self.vendor = part.vendor
        self.sku = part.sku
        self.count_per_sku = part.count_per_sku

        self.material = await part.get_mcftt("material")
        self.color = await part.get_mcftt("color")
        self.finish = await part.get_mcftt("finish")

    def compose(self):
        result = {
            "name": self.name,
            "count": self.count,
            "material": self.material,
            "color": self.color,
            "finish": self.finish,
            "format": self.format,
        }
        if self.binary:
            result["binary"] = self.binary
        if self.vendor and self.sku:
            result["vendor"] = self.vendor
            result["sku"] = self.sku
            if self.count_per_sku:
                result["count_per_sku"] = self.count_per_sku
            else:
                result["count_per_sku"] = 1
        return result

    def add_binary(self, format: str, binary: bytes):
        self.format = format
        self.binary = binary

    def __repr__(self):
        result = self.name + "#"
        result += str(self.count)
        return result


class ProviderCart:
    """Describes a cart of parts"""

    # TODO(clairbee): add a lock

    parts: dict[str, ProviderCartItem]
    qos: str = None

    def __init__(self, qos: str = None):
        self.parts = {}
        self.qos = qos

    async def add_objects(self, ctx, objects: list[str]):
        """Add parts or assemblies"""
        for object in objects:
            await self.add_object(ctx, object)

    async def add_object(self, ctx, object: str):
        """Add a part or an assembly"""
        # The input object may be an assembly
        # But self.cart must contain parts only
        name, count = resolve_cart_item(object)
        project_name, object_name = resolve_resource_path(
            ctx.current_project_path,
            name,
        )
        prj = ctx.get_project(project_name)

        part = prj.get_part(object_name, quiet=True)
        if part:
            pc_logging.debug(f"Adding part '{object_name}' to the cart")
            item = ProviderCartItem()
            await item.set_spec(ctx, name)
            self.add_item(item, count)
        else:
            assembly = prj.get_assembly(object_name)
            if assembly:
                pc_logging.debug(f"Adding assembly '{object_name}' to the cart")
                bom = await assembly.get_bom()
                tasks = []
                for part_name, part_count in bom.items():
                    pc_logging.debug(f"Adding part '{part_name}' to the cart")
                    part_spec = part_name + "#" + str(part_count)
                    tasks.append(
                        asyncio.create_task(
                            self.add_item_from_spec(ctx, part_spec, count)
                        )
                    )
                await asyncio.gather(*tasks)
            else:
                raise Exception(
                    f"Part or assembly '{object_name}' not found in project '{project_name}'"
                )

    async def add_item_from_spec(self, ctx, part_spec: str, count=1):
        part_item = ProviderCartItem()
        await part_item.set_spec(ctx, part_spec)
        self.add_item(part_item, count)

    async def add_part_specs(self, ctx, part_specs: list[str]):
        """Add parts"""
        for part_spec in part_specs:
            await self.add_part_spec(ctx, part_spec)

    async def add_part_spec(self, ctx, part_spec: str):
        """Add a part"""
        item = ProviderCartItem()
        await item.set_spec(ctx, part_spec)
        return self._add_item(item)

    def add_item(self, item: ProviderCartItem, count=1):
        """Copy the cart item into this cart in a way that the original cart is not affected"""
        new_item = copy.copy(item)
        self._add_item(new_item, count)
        return new_item

    def _add_item(self, item: ProviderCartItem, count=1):
        """Add item without 'copy.copy()'"""
        assert item.count > 0
        item.count *= count
        if item.name in self.parts:
            self.parts[item.name].count += item.count
        else:
            self.parts[item.name] = item
        return item

    def compose(self):
        req = {"parts": {}, "qos": self.qos}

        for name, part in self.parts.items():
            req["parts"][name] = part.compose()

        return req

    def __repr__(self):
        return str(self.compose())



#
# OpenVMP, 2023-2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-01
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within the python sandbox environment (python runtime)
# to invoke `build1213d` scripts.

from collections.abc import Iterable
import os
import re
import sys

from cadquery import cqgi
import build123d

sys.path.append(os.path.dirname(__file__))
import wrapper_common
import py_stubs.ocp_vscode  # Make 'sys.modules["py_stubs.ocp_vscode"]' available
from ocp_tessellate.ocp_utils import (
    get_downcasted_shape,
    downcast,
    is_build123d,
    is_build123d_compound,
    is_build123d_shell,
    is_wrapped,
    is_vector,
    is_topods_compound,
    is_topods_shape,
    vertex,
)


def process(path, request):
    build_parameters = {}
    if "patch" in request:
        patch = request["patch"]
    else:
        patch = {}
    if "build_parameters" in request:
        build_parameters = request["build_parameters"]

    script = open(path, "r").read()
    for old, new in patch.items():
        script = re.sub(old, new, script, flags=re.MULTILINE)

    if "import partcad" in script:
        script = (
            "import logging\nlogging.basicConfig(level=60)\n"  # Disable PartCAD logging
            + script
        )

    # Ignore ocp_vscode as it is of no use in the sandboxed environment
    # and it produces a lot of sporadic output to stdout and stderr
    sys.modules["ocp_vscode"] = sys.modules["py_stubs.ocp_vscode"]
    if "from ocp_vscode import " in script:
        script = re.sub(
            r"(from ocp_vscode import .*)\n",
            "saved_show_object_early=show_object\n\\1\nimport ocp_vscode\nocp_vscode.saved_show_object=saved_show_object_early\n",
            script,
        )
    if "import ocp_vscode" in script:
        script = script.replace(
            "import ocp_vscode",
            "import ocp_vscode\nocp_vscode.saved_show_object=show_object\n#",
        )

    # Execute the script
    script_object = cqgi.parse(script)
    build_result = script_object.build(build_parameters=build_parameters)

    if not build_result.success:
        wrapper_common.handle_exception(build_result.exception, path)

    results = list()
    # TODO(clairbee): make it recursive to handle nested lists and unify handling of items on different levels of nesting
    for result in build_result.results:
        shape = result.shape
        converted = list()

        # BuildPart, BuildSketch, BuildLine
        if is_build123d(shape):
            shape = getattr(shape, shape._obj_name)  # convert to direct API

        # TODO(clairbee): do we really want to explode compounds?
        if is_build123d_compound(shape):
            converted.append(get_downcasted_shape(shape.wrapped))

        elif is_build123d_shell(shape):
            faces = []
            for face in shape.faces():
                faces.append(get_downcasted_shape(face.wrapped))
            converted.extend(faces)

        # TODO(clairbee): is this needed?
        # elif is_build123d_shape(shape):
        #     converted.append(get_downcasted_shape(shape.wrapped))

        # TODO(clairbee): migrate the below to cadquery
        # elif is_cadquery_sketch(shape):
        #     converted.extend(conv_sketch(shape))

        # elif is_cadquery(shape):
        #     for v in shape.vals():
        #         if is_cadquery_sketch(v):
        #             # obj = conv_sketch(v)
        #             pass
        #         elif is_vector(v):
        #             obj = [vertex(v.wrapped)]
        #             converted.extend(obj)
        #         else:
        #             obj = [v.wrapped]
        #             converted.extend(obj)

        elif is_wrapped(shape):
            if is_vector(shape):
                converted.append(vertex(shape.wrapped))
            elif is_topods_shape(shape):
                converted.extend(
                    get_downcasted_shape(shape.wrapped)
                )  # TODO(clairbee): append(shape.wrapped)? append(downcast(shape.wrapped))?
            else:
                converted.append(shape.wrapped)

        elif isinstance(shape, str):
            converted.append(shape)

        elif isinstance(shape, Iterable):

            def unpack(objs):
                for obj in objs:
                    if is_wrapped(obj):
                        if is_vector(obj):
                            yield vertex(obj.wrapped)
                        else:
                            yield downcast(obj.wrapped)
                    else:
                        yield obj

            converted.append(list(unpack(shape)))

        elif is_topods_compound(shape):
            converted.append(shape)

        elif is_topods_shape(shape):
            converted.append(downcast(shape))

        else:
            sys.stderr.write(f"Unknown object type: {type(shape)}\n")
            converted.append(shape)

        # TODO(clairbee): is this needed?
        # if len(converted) > 0 and is_compound_list(converted):
        #     converted = get_downcasted_shape(converted[0])

        def solidify(obj):
            if hasattr(obj, "solid"):
                try:
                    obj = obj.solid().wrapped
                except Exception as e:
                    pass
            return obj

        converted = list(
            map(
                solidify,
                converted,
            )
        )
        results.extend(converted)

    return {
        "success": build_result.success,
        "exception": build_result.exception,
        "shapes": results,
    }


path, request = wrapper_common.handle_input()

# Call build123d through CQGI
model = process(path, request)

wrapper_common.handle_output(model)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

from .provider_request import ProviderRequest


class ProviderRequestCaps(ProviderRequest):
    def __init__(self):
        pass

    def compose(self):
        # This request has no parameters
        return {}



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-04-17
#
# Licensed under Apache License, Version 2.0.
#

import typing


class FileFactory:
    path: typing.Optional[str] = None

    def __init__(self, ctx, source_project, target_project, config):
        self.config = config
        self.ctx = ctx
        self.project = source_project

    async def download(self, path):
        raise NotImplementedError(
            "FileFactory.download is implemented in child classes"
        )



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import copy
import typing

from . import sketch_config
from . import sketch_factory as pf
from . import logging as pc_logging
from .utils import resolve_resource_path


class SketchFactoryEnrich(pf.SketchFactory):
    source_sketch_name: str
    source_project_name: typing.Optional[str]
    source: str

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "InitEnrich", target_project.name, config["name"]
        ):
            super().__init__(ctx, source_project, target_project, config)

            # Determine the sketch the 'enrich' points to
            if "source" in config:
                self.source_sketch_name = config["source"]
            else:
                self.source_sketch_name = config["name"]
                if not "project" in config:
                    raise Exception(
                        "Enrich needs either the source sketch name or the source project name"
                    )

            if "project" in config:
                self.source_project_name = config["project"]
                if (
                    self.source_project_name == "this"
                    or self.source_project_name == ""
                ):
                    self.source_project_name = source_project.name
            else:
                if ":" in self.source_sketch_name:
                    self.source_project_name, self.source_sketch_name = (
                        resolve_resource_path(
                            source_project.name,
                            self.source_sketch_name,
                        )
                    )
                else:
                    self.source_project_name = source_project.name
            self.source = (
                self.source_project_name + ":" + self.source_sketch_name
            )

            pc_logging.debug("Initializing an enrich to %s" % self.source)

            # Get the config of the sketch the 'enrich' points to
            orig_source_project = source_project
            if self.source_project_name == source_project.name:
                augmented_config = source_project.get_sketch_config(
                    self.source_sketch_name
                )
            else:
                source_project = ctx.get_project(self.source_project_name)
                augmented_config = source_project.get_sketch_config(
                    self.source_sketch_name
                )
            if augmented_config is None:
                pc_logging.error(
                    "Failed to find the sketch to enrich: %s"
                    % self.source_sketch_name
                )
                return

            augmented_config = copy.deepcopy(augmented_config)
            # TODO(clairbee): ideally whatever we pull from the project is already normalized
            augmented_config = sketch_config.SketchConfiguration.normalize(
                self.source_sketch_name,
                augmented_config,
            )

            # Drop fields we don't want to be inherited by enriched clones
            # TODO(clairbee): keep aliases if they are a function of the orignal name
            if "aliases" in augmented_config:
                del augmented_config["aliases"]

            # Fill in all non-enrich-specific properties from the enrich config into
            # the original config
            for prop_to_copy in config:
                if (
                    prop_to_copy == "type"
                    or prop_to_copy == "path"
                    or prop_to_copy == "orig_name"
                    or prop_to_copy == "source"
                    or prop_to_copy == "project"
                    or prop_to_copy == "with"
                ):
                    continue
                augmented_config[prop_to_copy] = config[prop_to_copy]

            # Fill in the parameter values using the simplified "with" option
            if "with" in config:
                for param in config["with"]:
                    augmented_config["parameters"][param]["default"] = config[
                        "with"
                    ][param]
            orig_source_project.init_sketch_by_config(
                augmented_config,
                source_project,
            )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#


import os

from .assembly_factory import AssemblyFactory
from . import logging as pc_logging


class AssemblyFactoryFile(AssemblyFactory):
    def __init__(
        self, ctx, source_project, target_project, config, extension=""
    ):
        super().__init__(ctx, source_project, target_project, config)

        if "path" in config:
            self.path = config["path"]
        else:
            self.path = self.orig_name + extension

        if not os.path.isdir(source_project.config_dir):
            raise Exception(
                "ERROR: The project config directory must be a directory, found: '%s'"
                % source_project.config_dir
            )
        self.path = os.path.join(source_project.config_dir, self.path)
        if not os.path.exists(self.path):
            raise Exception("ERROR: The assembly path must exist")

    async def instantiate(self, assembly):
        if not self.fileFactory is None and not os.path.exists(assembly.path):
            with pc_logging.Action(
                "File", self.target_project.name, assembly.name
            ):
                await self.fileFactory.download(assembly.path)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import typing

from .shape_ai import ShapeWithAi
from . import sync_threads as pc_thread


class Sketch(ShapeWithAi):
    path: typing.Optional[str] = None

    def __init__(self, config: object = {}):
        super().__init__(config)

        self.kind = "sketches"

    async def get_shape(self):
        async with self.lock:
            if self.shape is None:
                self.shape = await pc_thread.run_async(self.instantiate, self)
            return self.shape

    def ref_inc(self):
        # Not applicable to sketches
        # TODO(clairbee): move reference counter from Shape to another class
        #                 that would be used by both Part and Assembly
        pass

    def clone(self):
        # Not applicable to sketches
        # TODO(clairbee): move clone() from Shape to another class
        #                 that would be used by both Part and Assembly
        pass



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-04-17
#
# Licensed under Apache License, Version 2.0.
#

import aiofiles
import aiohttp
import os

from .file_factory import FileFactory
from .logging import debug


class FileFactoryUrl(FileFactory):
    url: str = None

    def __init__(self, ctx, source_project, target_project, config):
        super().__init__(ctx, source_project, target_project, config)

        self.url = config["fileUrl"]

    async def download(self, path):
        debug("Downloading file from %s to %s" % (self.url, path))

        dirs = os.path.dirname(path)
        if dirs != "" and not os.path.exists(dirs):
            os.makedirs(dirs)

        async with aiohttp.ClientSession() as session:
            r = await session.get(self.url)
            content = await r.read()

        async with aiofiles.open(path, "wb") as f:
            await f.write(content)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from .interface import Interface


class Mating:
    """A mating between two interfaces."""

    source: Interface
    target: Interface
    desc: str
    count: int

    source_port_selector: str
    target_port_selector: str

    def __init__(
        self,
        source: Interface,
        target: Interface,
        config: dict = {},
        reverse: bool = False,
    ):
        if config is None:
            config = {}
        elif isinstance(config, str):
            config = {config: None}
        elif isinstance(config, list):
            config = {c: None for c in config}
        elif not isinstance(config, dict):
            raise ValueError("Invalid mating configuration")

        self.source = source
        self.target = target
        self.desc = config["desc"] if "desc" in config else ""
        self.count = 0

        if "sourcePortSelector" in config:
            if reverse:
                self.target_port_selector = config["sourcePortSelector"]
            else:
                self.source_port_selector = config["sourcePortSelector"]
        else:
            self.source_port_selector = None

        if "targetPortSelector" in config:
            if reverse:
                self.source_port_selector = config["targetPortSelector"]
            else:
                self.target_port_selector = config["targetPortSelector"]
        else:
            self.target_port_selector = None



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from .sketch_factory_file import SketchFactoryFile
from .runtime_python import PythonRuntime


# TODO(clairbee): create ShapeFactoryPython to be reused
#                 by corresponding Sketch, Part and Assembly factories
class SketchFactoryPython(SketchFactoryFile):
    runtime: PythonRuntime
    cwd: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config,
        can_create=False,
        python_version=None,
    ):
        super().__init__(
            ctx,
            source_project,
            target_project,
            config,
            extension=".py",
            can_create=can_create,
        )
        self.cwd = config.get("cwd", None)

        if python_version is None:
            # TODO(clairbee): stick to a default constant or configured version
            python_version = self.project.python_version
        self.runtime = self.ctx.get_python_runtime(python_version)

    async def prepare_python(self):
        """
        This method is called by child classes
        to prepare the Python environment
        before instantiating the sketch.
        """

        # Install dependencies of this package
        await self.runtime.prepare_for_package(self.project)
        await self.runtime.prepare_for_shape(self.config)

    def info(self, sketch):
        info: dict[str, object] = sketch.shape_info()
        info.update(
            {
                "runtime_version": self.runtime.version,
                "runtime_path": self.runtime.path,
            }
        )
        return info

    async def instantiate(self, sketch):
        await super().instantiate(sketch)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

from .part_factory_scad import PartFactoryScad
from .part_factory_feature_ai import PartFactoryFeatureAi
from . import logging as pc_logging


class PartFactoryAiScad(PartFactoryScad, PartFactoryFeatureAi):
    def __init__(self, ctx, source_project, target_project, config):
        # Override the path determined by the parent class to enable "enrich"
        config["path"] = config["name"] + ".scad"

        with pc_logging.Action(
            "InitAiScad", target_project.name, config["name"]
        ):
            PartFactoryFeatureAi.__init__(
                self,
                config,
                "scad",
                "OpenSCAD script",
                """Generate a complete functioning script, not just a code snippet.
There are no other non-standard modules available to import.
Define all necessary functions and constants.
Do not generate comments.
Do not export anything.
""",
            )
            PartFactoryScad.__init__(
                self,
                ctx,
                source_project,
                target_project,
                config,
                can_create=True,
            )

            self.on_init_ai()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from .shape_config import ShapeConfiguration


class SketchConfiguration(ShapeConfiguration):
    def __init__(self, name, config):
        super().__init__(name, config)

    @staticmethod
    def normalize(name, config):
        if isinstance(config, str):
            # This is a short form alias
            config = {"type": "alias", "source": config}

        if "parameters" in config:
            for param_name, param_value in config["parameters"].items():
                # Expand short formats
                if isinstance(param_value, str):
                    config["parameters"][param_name] = {
                        "type": "string",
                        "default": param_value,
                    }
                elif isinstance(param_value, bool):
                    config["parameters"][param_name] = {
                        "type": "bool",
                        "default": param_value,
                    }
                elif isinstance(param_value, float):
                    config["parameters"][param_name] = {
                        "type": "float",
                        "default": param_value,
                    }
                elif isinstance(param_value, int):
                    config["parameters"][param_name] = {
                        "type": "int",
                        "default": param_value,
                    }
                elif isinstance(param_value, list):
                    config["parameters"][param_name] = {
                        "type": "array",
                        "default": param_value,
                    }
                # All params are float unless another type is explicitly speciifed
                elif (
                    isinstance(param_value, dict) and not "type" in param_value
                ):
                    param_value["type"] = "float"

        return ShapeConfiguration.normalize(name, config)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import math

from OCP.TopLoc import TopLoc_Location
from OCP.gp import (
    gp_Trsf,
    gp_Ax1,
    gp_Pnt,
    gp_Vec,
    gp_Dir,
)

from . import logging as pc_logging


class Location:
    wrapped: TopLoc_Location

    def __init__(self, arg=None):
        if arg is None:
            self.wrapped = TopLoc_Location()
        elif isinstance(arg, gp_Trsf):
            self.wrapped = TopLoc_Location(arg)
        elif isinstance(arg, list):
            trsf = Location.trsf_from_coords(arg[0], arg[1], arg[2])
            self.wrapped = TopLoc_Location(trsf)
        else:
            raise ValueError("geom.Location: Invalid argument type")

    def __repr__(self):
        trsf = self.wrapped.Transformation()
        trans = trsf.TranslationPart()
        quat = trsf.GetRotation()
        dir = gp_Vec()
        angle = quat.GetVectorAndAngle(dir)
        angle = 180.0 * angle[0] / math.pi
        return f"Location({[trans.X(), trans.Y(), trans.Z()]}, {[dir.X(), dir.Y(), dir.Z()]}, {angle})"

    @classmethod
    def trsf_from_coords(self, t: list[float], ax: list[float], angle: float):
        # pc_logging.info(f"trsf_from_coords: {t}, {ax}, {angle}")
        T = gp_Trsf()
        T.SetRotation(
            gp_Ax1(gp_Pnt(), gp_Dir(ax[0], ax[1], ax[2])),
            angle * math.pi / 180.0,
        )
        T.SetTranslationPart(gp_Vec(t[0], t[1], t[2]))
        return T



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.


class PluginExportPng:
    """This is a stub for plugins that provide functionality of exporting to PNG"""

    def is_supported(self):
        return False

    def export(self, project, svg_path, width, height, filepath):
        return


plugin_export_png = PluginExportPng()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

from .provider_request import ProviderRequest


class ProviderRequestAvail(ProviderRequest):
    name: str
    vendor: str
    sku: str
    count_per_sku: int
    count: int

    def __init__(
        self, name: str, vendor: str, sku: str, count_per_sku: int, count: int
    ):
        self.name = name
        self.vendor = vendor
        self.sku = sku
        self.count_per_sku = count_per_sku
        self.count = count

    def compose(self):
        return {
            "name": self.name,
            "vendor": self.vendor,
            "sku": self.sku,
            "count_per_sku": self.count_per_sku,
            "count": self.count,
        }



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

from jinja2 import Environment, FileSystemLoader, ChoiceLoader
import json
import os
from packaging.specifiers import SpecifierSet
import sys
import yaml
import math

from . import consts
from . import logging as pc_logging
from . import exception as pc_exception

DEFAULT_CONFIG_FILENAME = "partcad.yaml"


class Configuration:
    name: str

    def __init__(
        self, name, config_path=DEFAULT_CONFIG_FILENAME, include_paths=[]
    ):
        self.name = name
        self.config_obj = {}
        self.config_dir = config_path
        self.config_path = config_path
        self.broken = False

        if os.path.isdir(config_path):
            self.config_path = os.path.join(
                config_path, DEFAULT_CONFIG_FILENAME
            )
        else:
            self.config_dir = os.path.dirname(os.path.abspath(config_path))

        if not os.path.isfile(self.config_path):
            pc_logging.error(
                "PartCAD configuration file is not found: '%s'"
                % self.config_path
            )
            self.broken = True
            return

        # Read the body of the configuration file
        fp = open(self.config_path, "r")
        config = fp.read()
        fp.close()

        # Resolve Jinja templates
        loaders = [FileSystemLoader(self.config_dir + os.path.sep)]
        # TODO(clairbee): mark the build as non-hermetic if includePaths is used
        for include_path in include_paths:
            include_path = (
                os.path.join(self.config_dir, include_path) + os.path.sep
            )
            loaders.append(FileSystemLoader(include_path))
        loader = ChoiceLoader(loaders)
        template = Environment(loader=loader).from_string(config)
        config = template.render(
            {
                "package_name": name,
                "M_PI": math.pi,
                "PI": math.pi,
                "SQRT_2": math.sqrt(2),
                "SQRT_3": math.sqrt(3),
                "SQRT_5": math.sqrt(5),
                "INCH": 25.4,
                "INCHES": 25.4,
                "FOOT": 304.8,
                "FEET": 304.8,
            }
        )

        # Parse the config
        if self.config_path.endswith(".yaml"):
            self.config_obj = yaml.safe_load(config)
        if self.config_path.endswith(".json"):
            self.config_obj = json.load(config)

        if self.config_obj is None:
            self.config_obj = {}

        if name == consts.ROOT and "name" in self.config_obj:
            name = self.config_obj["name"]
            self.name = name
        else:
            self.config_obj["name"] = name

        if not "render" in self.config_obj or self.config_obj["render"] is None:
            self.config_obj["render"] = {}

        # option: "partcad"
        # description: the version of PartCAD required to handle this package
        # values: string initializer for packaging.specifiers.SpecifierSet
        # default: None
        if "partcad" in self.config_obj:
            partcad_requirements = SpecifierSet(self.config_obj["partcad"])
            partcad_version = sys.modules["partcad"].__version__
            if partcad_version not in partcad_requirements:
                # TODO(clairbee): add better error and exception handling
                raise pc_exception.NeedsUpdateException(
                    "ERROR: Incompatible PartCAD version! %s does not satisfy %s"
                    % (partcad_version, partcad_requirements)
                )

        # option: "pythonVersion"
        # description: the version of python to use in sandboxed environments if any
        # values: string (e.g. "3.10")
        # default: <The major and minor version of the current interpreter>
        if "pythonVersion" == self.config_obj:
            self.python_version = self.config_obj["pythonVersion"]
        else:
            self.python_version = "%d.%d" % (
                sys.version_info.major,
                sys.version_info.minor,
            )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import math
import typing

from .shape_ai import ShapeWithAi
from . import sync_threads as pc_thread
from . import logging as pc_logging


class Part(ShapeWithAi):
    path: typing.Optional[str] = None
    vendor: typing.Optional[str] = None
    sku: typing.Optional[str] = None
    url: typing.Optional[str] = None
    count_per_sku: int = None
    count: int = None

    def __init__(self, config: object = {}, shape=None):
        super().__init__(config)

        self.kind = "parts"
        self.shape = shape

        self.vendor = None
        if "vendor" in config:
            self.vendor = config["vendor"]
        self.sku = None
        if "sku" in config:
            self.sku = config["sku"]
        self.url = None
        if "url" in config:
            self.url = config["url"]
        if "count_per_sku" in config:
            self.count_per_sku = int(config["count_per_sku"])
        else:
            self.count_per_sku = 1
        self.count = 0

    async def get_shape(self):
        async with self.lock:
            if self.shape is None:
                self.shape = await pc_thread.run_async(self.instantiate, self)
            return self.shape

    def ref_inc(self):
        # TODO(clairbee): add a thread lock here
        self.count += 1

    def clone(self):
        cloned = Part(self.name, self.config, self.shape)
        cloned.count = self.count
        return cloned

    async def get_mcftt(self, property: str):
        # Get the material, color, finish, texture or tolerance of the part
        if (
            self.vendor is None
            and self.sku is None
            and (
                not "parameters" in self.config
                or not property in self.config["parameters"]
            )
        ):
            shape = await self.get_shape()
            # TODO(clairbee): derive the property from the model

            if property == "finish":
                # By default, the finish is set to "none"
                value = "none"
            else:
                # By default, the parameter is not set
                value = None

            if value is not None:
                if "parameters" not in self.config:
                    self.config["parameters"] = {}
                self.config["parameters"][property] = {
                    "type": "string",
                    "enum": [value],
                    "default": value,
                }
            else:
                pc_logging.warning(f"Part '{self.name}' has no '{property}'")

            return value

        if (
            "parameters" not in self.config
            or property not in self.config["parameters"]
            or "default" not in self.config["parameters"][property]
        ):
            return None
        return self.config["parameters"][property]["default"]

    def _render_txt_real(self, file):
        file.write(self.name + ": " + self.count + "\n")

    def _render_markdown_real(self, file):
        name = self.name
        vendor = ""
        sku = ""
        if not self.desc is None:
            name = self.desc
        if not self.vendor is None:
            vendor = self.vendor
        if not self.sku is None:
            sku = self.sku
        if self.url is None:
            label = name
        else:
            label = "[" + name + "](" + self.url + ")"
            sku = "[" + sku + "](" + self.url + ")"
        count = str(math.ceil(self.count / self.count_per_sku))
        img_url = self._get_svg_url()

        file.write(
            "| "
            + label
            + " | "
            + count
            + " |"
            + vendor
            + " |"
            + sku
            + " |"
            + " !["
            + name
            + "]("
            + img_url
            + ")"
            + " |\n"
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import copy
import os

# from pprint import pformat
import ruamel.yaml
import threading
import typing

from . import consts
from . import factory
from . import logging as pc_logging
from . import project_config
from . import interface
from . import sketch
from . import sketch_config
from .sketch_factory_alias import SketchFactoryAlias
from .sketch_factory_enrich import SketchFactoryEnrich
from .sketch_factory_basic import SketchFactoryBasic
from .sketch_factory_dxf import SketchFactoryDxf
from .sketch_factory_svg import SketchFactorySvg
from .sketch_factory_build123d import SketchFactoryBuild123d
from .sketch_factory_cadquery import SketchFactoryCadquery
from . import part
from . import part_config
from .part_factory_extrude import PartFactoryExtrude
from . import part_factory_scad as pfscad
from . import part_factory_step as pfs
from . import part_factory_stl as pfstl
from . import part_factory_3mf as pf3
from .part_factory_ai_cadquery import PartFactoryAiCadquery
from .part_factory_ai_build123d import PartFactoryAiBuild123d
from .part_factory_ai_openscad import PartFactoryAiScad
from . import part_factory_cadquery as pfc
from . import part_factory_build123d as pfb
from . import part_factory_alias as pfa
from . import part_factory_enrich as pfe
from . import assembly
from . import assembly_config
from . import provider
from . import provider_config
from .render import render_cfg_merge
from .utils import resolve_resource_path, normalize_resource_path


class Project(project_config.Configuration):

    class InterfaceLock(object):
        def __init__(self, prj, interface_name: str):
            prj.interface_locks_lock.acquire()
            if not interface_name in prj.interface_locks:
                prj.interface_locks[interface_name] = threading.Lock()
            self.lock = prj.interface_locks[interface_name]
            prj.interface_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    class SketchLock(object):
        def __init__(self, prj, sketch_name: str):
            prj.sketch_locks_lock.acquire()
            if not sketch_name in prj.sketch_locks:
                prj.sketch_locks[sketch_name] = threading.Lock()
            self.lock = prj.sketch_locks[sketch_name]
            prj.sketch_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    class PartLock(object):
        def __init__(self, prj, part_name: str):
            prj.part_locks_lock.acquire()
            if not part_name in prj.part_locks:
                prj.part_locks[part_name] = threading.Lock()
            self.lock = prj.part_locks[part_name]
            prj.part_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    class AssemblyLock(object):
        def __init__(self, prj, assembly_name: str):
            prj.assembly_locks_lock.acquire()
            if not assembly_name in prj.assembly_locks:
                prj.assembly_locks[assembly_name] = threading.Lock()
            self.lock = prj.assembly_locks[assembly_name]
            prj.assembly_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    class ProviderLock(object):
        def __init__(self, prj, provider_name: str):
            prj.provider_locks_lock.acquire()
            if not provider_name in prj.provider_locks:
                prj.provider_locks[provider_name] = threading.Lock()
            self.lock = prj.provider_locks[provider_name]
            prj.provider_locks_lock.release()

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    def __init__(self, ctx, name, path, include_paths=[]):
        super().__init__(name, path, include_paths)
        self.ctx = ctx

        # Protect the critical sections from access in different threads
        self.lock = threading.Lock()

        # The 'path' parameter is the config filename or the directory
        # where 'partcad.yaml' is present.
        # 'self.path' has to be set to the directory name.
        dir_name = path
        if not os.path.isdir(dir_name):
            dir_name = os.path.dirname(os.path.abspath(dir_name))
        self.path = dir_name

        # self.interface_configs contains the configs of all the interfaces in this project
        if (
            "interfaces" in self.config_obj
            and not self.config_obj["interfaces"] is None
        ):
            self.interface_configs = self.config_obj["interfaces"]
            # pc_logging.debug(
            #     "Interfaces: %s" % str(self.interface_configs.keys())
            # )
        else:
            self.interface_configs = {}
        # self.interfaces contains all the initialized interfaces in this project
        self.interfaces = {}
        self.interface_locks = {}
        self.interface_locks_lock = threading.Lock()

        # self.sketch_configs contains the configs of all the sketches in this project
        if (
            "sketches" in self.config_obj
            and not self.config_obj["sketches"] is None
        ):
            self.sketch_configs = self.config_obj["sketches"]
        else:
            self.sketch_configs = {}
        # self.sketches contains all the initialized sketches in this project
        self.sketches = {}
        self.sketch_locks = {}
        self.sketch_locks_lock = threading.Lock()

        # self.part_configs contains the configs of all the parts in this project
        if "parts" in self.config_obj and not self.config_obj["parts"] is None:
            self.part_configs = self.config_obj["parts"]
        else:
            self.part_configs = {}
        # self.parts contains all the initialized parts in this project
        self.parts = {}
        self.part_locks = {}
        self.part_locks_lock = threading.Lock()

        # self.assembly_configs contains the configs of all the assemblies in this project
        if (
            "assemblies" in self.config_obj
            and not self.config_obj["assemblies"] is None
        ):
            self.assembly_configs = self.config_obj["assemblies"]
        else:
            self.assembly_configs = {}
        # self.assemblies contains all the initialized assemblies in this project
        self.assemblies = {}
        self.assembly_locks = {}
        self.assembly_locks_lock = threading.Lock()

        # self.provider_configs contains the configs of all the providers in this project
        if (
            "providers" in self.config_obj
            and not self.config_obj["providers"] is None
        ):
            self.provider_configs = self.config_obj["providers"]
        else:
            self.provider_configs = {}
        # self.providers contains all the initialized providers in this project
        self.providers = {}
        self.provider_locks = {}
        self.provider_locks_lock = threading.Lock()

        if (
            "desc" in self.config_obj
            and not self.config_obj["desc"] is None
            and isinstance(self.config_obj["desc"], str)
        ):
            self.desc = self.config_obj["desc"].strip()
        else:
            self.desc = ""

        self.init_sketches()
        self.init_interfaces()  # After sketches
        self.init_mates()  # After interfaces
        self.init_parts()  # After sketches and interfaces, and mates
        self.init_assemblies()  # after parts
        self.init_providers()  # after parts
        self.init_suppliers()  # after suppliers

    # TODO(clairbee): Implement get_cover()
    # def get_cover(self):
    #     if not "cover" in self.config_obj or self.config_obj["cover"] is None:
    #         return None
    #     if isinstance(self.config_obj["cover"], str):
    #         return os.path.join(self.config_dir, self.config_obj["cover"])
    #     elif "package" in self.config_obj["cover"]:
    #         return self.ctx.get_project(
    #             self.path + "/" + self.config_obj["cover"]["package"]
    #         ).get_cover()

    def get_child_project_names(self):
        if self.broken:
            pc_logging.info("Ignoring the broken package: %s" % self.name)
            return

        children = list()
        subfolders = [f.name for f in os.scandir(self.config_dir) if f.is_dir()]
        for subdir in list(subfolders):
            if os.path.exists(
                os.path.join(
                    self.config_dir,
                    subdir,
                    consts.DEFAULT_PACKAGE_CONFIG,
                )
            ):
                children.append(self.name + "/" + subdir)

        if (
            "import" in self.config_obj
            and not self.config_obj["import"] is None
        ):
            imports = self.config_obj["import"]
            if not self.config_obj.get("isRoot", False):
                filtered = filter(
                    lambda x: "onlyInRoot" not in imports[x]
                    or not imports[x]["onlyInRoot"],
                    imports,
                )
                imports = list(filtered)
            children.extend(
                list(
                    map(
                        lambda project_name: self.name + "/" + project_name,
                        imports,
                    )
                )
            )
        return children

    def init_mates(self):
        mates = self.config_obj.get("mates", {})
        for source_interface_name, mate_config in mates.items():
            if not ":" in source_interface_name:
                source_interface_name = self.name + ":" + source_interface_name
            source_package_name, short_source_interface_name = (
                resolve_resource_path(self.name, source_interface_name)
            )

            # Short-circut the case when the source package is the current one
            # to avoid recursive package loading
            if source_package_name == self.name:
                source_package = self
            else:
                source_package = self.ctx.get_project(source_package_name)

            source_interface = source_package.get_interface(
                short_source_interface_name
            )
            if source_interface is None:
                raise Exception(
                    "Failed to find the source interface to mate: %s"
                    % source_interface_name
                )
            source_interface.add_mates(self, mate_config)

    def get_interface_config(self, interface_name):
        if not interface_name in self.interface_configs:
            return None
        return self.interface_configs[interface_name]

    def init_interfaces(self):
        if self.interface_configs is None:
            return

        for interface_name in self.interface_configs.keys():
            config = self.get_interface_config(interface_name)
            config["name"] = interface_name
            self.init_interface_by_config(config)

    def init_interface_by_config(self, config, source_project=None):
        if source_project is None:
            source_project = self

        interface_name: str = config["name"]
        self.interfaces[interface_name] = interface.Interface(
            interface_name, source_project, config
        )

    def get_interface(self, interface_name) -> interface.Interface:
        self.lock.acquire()

        # See if it's already available
        if (
            interface_name in self.interfaces
            and not self.interfaces[interface_name] is None
        ):
            p = self.interfaces[interface_name]
            self.lock.release()
            return p

        with Project.InterfaceLock(self, interface_name):
            # Release the project lock, and continue with holding the interface lock only
            self.lock.release()

            # This is just a regular interface name, no params (interface_name == result_name)
            if not interface_name in self.interface_configs:
                # We don't know anything about such a interface
                pc_logging.error(
                    "Interface '%s' not found in '%s'",
                    interface_name,
                    self.name,
                )
                return None
            # This is not yet created (invalidated?)
            config = self.get_interface_config(interface_name)
            config["name"] = interface_name
            self.init_interface_by_config(config)
            return self.interfaces[interface_name]

    def get_sketch_config(self, sketch_name):
        if not sketch_name in self.sketch_configs:
            return None
        return self.sketch_configs[sketch_name]

    def init_sketches(self):
        if self.sketch_configs is None:
            return

        for sketch_name in self.sketch_configs:
            config = self.get_sketch_config(sketch_name)
            config = sketch_config.SketchConfiguration.normalize(
                sketch_name, config
            )
            self.init_sketch_by_config(config)

    def init_sketch_by_config(self, config, source_project=None):
        if source_project is None:
            source_project = self

        sketch_name: str = config["name"]

        if not "type" in config:
            raise Exception(
                "ERROR: Sketch type is not specified: %s: %s"
                % (sketch_name, config)
            )
        elif config["type"] == "build123d":
            SketchFactoryBuild123d(self.ctx, source_project, self, config)
        elif config["type"] == "cadquery":
            SketchFactoryCadquery(self.ctx, source_project, self, config)
        elif config["type"] == "dxf":
            SketchFactoryDxf(self.ctx, source_project, self, config)
        elif config["type"] == "svg":
            SketchFactorySvg(self.ctx, source_project, self, config)
        elif config["type"] == "basic":
            SketchFactoryBasic(self.ctx, source_project, self, config)
        elif config["type"] == "alias":
            SketchFactoryAlias(self.ctx, source_project, self, config)
        elif config["type"] == "enrich":
            SketchFactoryEnrich(self.ctx, source_project, self, config)
        else:
            pc_logging.error(
                "Invalid sketch type encountered: %s: %s"
                % (sketch_name, config)
            )
            return None

        # Initialize aliases if they are declared implicitly
        if "aliases" in config and not config["aliases"] is None:
            for alias in config["aliases"]:
                if ";" in sketch_name:
                    # Copy parameters
                    alias += sketch_name[sketch_name.index(";") :]
                alias_sketch_config = {
                    "type": "alias",
                    "name": alias,
                    "source": ":" + sketch_name,
                }
                alias_sketch_config = (
                    sketch_config.SketchConfiguration.normalize(
                        alias, alias_sketch_config
                    )
                )
                pfa.SketchFactoryAlias(
                    self.ctx, source_project, self, alias_sketch_config
                )

    def get_sketch(self, sketch_name, func_params=None) -> sketch.Sketch:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: dict[str, typing.Any] = {}
        if ";" in sketch_name:
            has_name_params = True
            base_sketch_name = sketch_name.split(";")[0]
            sketch_name_params_string = sketch_name.split(";")[1]

            for kv in sketch_name_params_string.split[","]:
                k, v = kv.split("")
                params[k] = v
        else:
            has_name_params = False
            base_sketch_name = sketch_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = sketch_name
        else:
            # Determine the name we want this parameterized sketch to have
            result_name = base_sketch_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if (
            result_name in self.sketches
            and not self.sketches[result_name] is None
        ):
            p = self.sketches[result_name]
            self.lock.release()
            return p

        with Project.SketchLock(self, result_name):
            # Release the project lock, and continue with holding the sketch lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular sketch name, no params (sketch_name == result_name)
                if not sketch_name in self.sketch_configs:
                    # We don't know anything about such a sketch
                    pc_logging.error(
                        "Sketch '%s' not found in '%s'", sketch_name, self.name
                    )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_sketch_config(sketch_name)
                config = sketch_config.SketchConfiguration.normalize(
                    sketch_name, config
                )
                self.init_sketch_by_config(config)

                if (
                    not sketch_name in self.sketches
                    or self.sketches[sketch_name] is None
                ):
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized sketch %s"
                        % sketch_name
                    )
                return self.sketches[sketch_name]

            # This sketch has params (sketch_name != result_name)
            if not base_sketch_name in self.sketches:
                pc_logging.error(
                    "Base sketch '%s' not found in '%s'",
                    base_sketch_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base sketch: %s" % base_sketch_name)

            # Now we have the original sketch name and the complete set of parameters
            config = self.sketch_configs[base_sketch_name]
            if config is None:
                pc_logging.error(
                    "The config for the base sketch '%s' is not found in '%s'",
                    base_sketch_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_sketch_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = sketch_config.SketchConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_sketch_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
                    elif config["parameters"][param_name]["type"] == "array":
                        config["parameters"][param_name]["default"] = param_value
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the sketch
            pc_logging.debug(
                "Initializing a parametrized sketch: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized sketch using the following config: %s"
            #     % pformat(config)
            # )
            self.init_sketch_by_config(config)

            # See if it worked
            if not result_name in self.sketches:
                pc_logging.error(
                    "Failed to instantiate parameterized sketch '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.sketches[result_name]

    def get_part_config(self, part_name):
        if not part_name in self.part_configs:
            return None
        return self.part_configs[part_name]

    def init_parts(self):
        if self.part_configs is None:
            return

        for part_name in self.part_configs:
            config = self.get_part_config(part_name)
            config = part_config.PartConfiguration.normalize(part_name, config)
            self.init_part_by_config(config)

    def init_part_by_config(self, config, source_project=None):
        if source_project is None:
            source_project = self

        part_name: str = config["name"]

        if not "type" in config:
            raise Exception(
                "ERROR: Part type is not specified: %s: %s"
                % (part_name, config)
            )
        elif config["type"] == "ai-cadquery":
            PartFactoryAiCadquery(self.ctx, source_project, self, config)
        elif config["type"] == "ai-build123d":
            PartFactoryAiBuild123d(self.ctx, source_project, self, config)
        elif config["type"] == "ai-openscad":
            PartFactoryAiScad(self.ctx, source_project, self, config)
        elif config["type"] == "cadquery":
            pfc.PartFactoryCadquery(self.ctx, source_project, self, config)
        elif config["type"] == "build123d":
            pfb.PartFactoryBuild123d(self.ctx, source_project, self, config)
        elif config["type"] == "step":
            pfs.PartFactoryStep(self.ctx, source_project, self, config)
        elif config["type"] == "stl":
            pfstl.PartFactoryStl(self.ctx, source_project, self, config)
        elif config["type"] == "3mf":
            pf3.PartFactory3mf(self.ctx, source_project, self, config)
        elif config["type"] == "scad":
            pfscad.PartFactoryScad(self.ctx, source_project, self, config)
        elif config["type"] == "extrude":
            PartFactoryExtrude(self.ctx, source_project, self, config)
        elif config["type"] == "alias":
            pfa.PartFactoryAlias(self.ctx, source_project, self, config)
        elif config["type"] == "enrich":
            pfe.PartFactoryEnrich(self.ctx, source_project, self, config)
        else:
            pc_logging.error(
                "Invalid part type encountered: %s: %s" % (part_name, config)
            )
            return None

        # Initialize aliases if they are declared implicitly
        if "aliases" in config and not config["aliases"] is None:
            for alias in config["aliases"]:
                if ";" in part_name:
                    # Copy parameters
                    alias += part_name[part_name.index(";") :]
                alias_part_config = {
                    "type": "alias",
                    "name": alias,
                    "source": ":" + part_name,
                }
                alias_part_config = part_config.PartConfiguration.normalize(
                    alias, alias_part_config
                )
                pfa.PartFactoryAlias(
                    self.ctx, source_project, self, alias_part_config
                )

    def get_part(self, part_name, func_params=None, quiet=False) -> part.Part:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: dict[str, typing.Any] = {}
        if ";" in part_name:
            has_name_params = True
            base_part_name = part_name.split(";")[0]
            part_name_params_string = part_name.split(";")[1]

            for kv in part_name_params_string.split(","):
                k, v = kv.split("=")
                params[k] = v
        else:
            has_name_params = False
            base_part_name = part_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = part_name
        else:
            # Determine the name we want this parameterized part to have
            result_name = base_part_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if result_name in self.parts and not self.parts[result_name] is None:
            p = self.parts[result_name]
            self.lock.release()
            return p

        with Project.PartLock(self, result_name):
            # Release the project lock, and continue with holding the part lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular part name, no params (part_name == result_name)
                if not part_name in self.part_configs:
                    # We don't know anything about such a part
                    if not quiet:
                        pc_logging.error(
                            "Part '%s' not found in '%s'", part_name, self.name
                        )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_part_config(part_name)
                config = part_config.PartConfiguration.normalize(
                    part_name, config
                )
                self.init_part_by_config(config)

                if not part_name in self.parts or self.parts[part_name] is None:
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized part %s"
                        % part_name
                    )
                return self.parts[part_name]

            # This part has params (part_name != result_name)
            if not base_part_name in self.parts:
                pc_logging.error(
                    "Base part '%s' not found in '%s'",
                    base_part_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base part: %s" % base_part_name)

            # Now we have the original part name and the complete set of parameters
            config = self.part_configs[base_part_name]
            if config is None:
                pc_logging.error(
                    "The config for the base part '%s' is not found in '%s'",
                    base_part_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_part_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = part_config.PartConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_part_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
                    elif config["parameters"][param_name]["type"] == "array":
                        config["parameters"][param_name]["default"] = param_value
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the part
            pc_logging.debug(
                "Initializing a parametrized part: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized part using the following config: %s"
            #     % pformat(config)
            # )
            self.init_part_by_config(config)

            # See if it worked
            if not result_name in self.parts:
                pc_logging.error(
                    "Failed to instantiate parameterized part '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.parts[result_name]

    def get_assembly_config(self, assembly_name):
        if not assembly_name in self.assembly_configs:
            return None
        return self.assembly_configs[assembly_name]

    def init_assemblies(self):
        if self.assembly_configs is None:
            return

        for assembly_name in self.assembly_configs:
            config = self.get_assembly_config(assembly_name)
            config = assembly_config.AssemblyConfiguration.normalize(
                assembly_name, config
            )
            factory.instantiate(
                "assembly", config["type"], self.ctx, self, self, config
            )

    def get_assembly(
        self, assembly_name, func_params=None
    ) -> assembly.Assembly:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: dict[str, typing.Any] = {}
        if ";" in assembly_name:
            has_name_params = True
            base_assembly_name = assembly_name.split(";")[0]
            assembly_name_params_string = assembly_name.split(";")[1]

            for kv in assembly_name_params_string.split(","):
                k, v = kv.split("=")
                params[k] = v
        else:
            has_name_params = False
            base_assembly_name = assembly_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = assembly_name
        else:
            # Determine the name we want this parameterized assembly to have
            result_name = base_assembly_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if (
            result_name in self.assemblies
            and not self.assemblies[result_name] is None
        ):
            p = self.assemblies[result_name]
            self.lock.release()
            return p

        with Project.AssemblyLock(self, result_name):
            # Release the project lock, and continue with holding the part lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular assembly name, no params (assembly_name == result_name)
                if not assembly_name in self.assembly_configs:
                    # We don't know anything about such an assembly
                    pc_logging.error(
                        "Assembly '%s' not found in '%s'",
                        assembly_name,
                        self.name,
                    )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_assembly_config(assembly_name)
                config = assembly_config.AssemblyConfiguration.normalize(
                    assembly_name, config
                )
                factory.instantiate(
                    "assembly", config["type"], self.ctx, self, self, config
                )

                if (
                    not assembly_name in self.assemblies
                    or self.assemblies[assembly_name] is None
                ):
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized assembly %s"
                        % assembly_name
                    )
                return self.assemblies[assembly_name]

            # This assembly has params (part_name != result_name)
            if not base_assembly_name in self.assemblies:
                pc_logging.error(
                    "Base assembly '%s' not found in '%s'",
                    base_assembly_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base assembly: %s" % base_assembly_name)

            # Now we have the original assembly name and the complete set of parameters
            config = self.assembly_configs[base_assembly_name]
            if config is None:
                pc_logging.error(
                    "The config for the base assembly '%s' is not found in '%s'",
                    base_assembly_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_assembly_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = assembly_config.AssemblyConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_assembly_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
                    elif config["parameters"][param_name]["type"] == "array":
                        config["parameters"][param_name]["default"] = param_value
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the assembly
            pc_logging.debug(
                "Initializing a parametrized assembly: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized assembly using the following config: %s"
            #     % pformat(config)
            # )
            factory.instantiate(
                "assembly", config["type"], self.ctx, self, self, config
            )

            # See if it worked
            if not result_name in self.assemblies:
                pc_logging.error(
                    "Failed to instantiate parameterized assembly '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.assemblies[result_name]

    def get_provider_config(self, provider_name):
        if not provider_name in self.provider_configs:
            return None
        return self.provider_configs[provider_name]

    def init_providers(self):
        if self.provider_configs is None:
            return

        for provider_name in self.provider_configs:
            config = self.get_provider_config(provider_name)
            config = provider_config.ProviderConfiguration.normalize(
                provider_name, config
            )
            factory.instantiate(
                "provider", config["type"], self.ctx, self, self, config
            )

    # TODO(clairbee): either call init_*_by_config or call
    #                  factory->instantite everywhere.
    #                  Recall what waas the thinking about it when the factory
    #                  class was introduced.

    def init_provider_by_config(self, config, source_project=None):
        if source_project is None:
            source_project = self
        factory.instantiate(
            "provider", config["type"], self.ctx, source_project, self, config
        )

    def get_provider(
        self, provider_name, func_params=None
    ) -> provider.Provider:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: dict[str, typing.Any] = {}
        if ";" in provider_name:
            has_name_params = True
            base_provider_name = provider_name.split(";")[0]
            provider_name_params_string = provider_name.split(";")[1]

            for kv in provider_name_params_string.split(","):
                k, v = kv.split("=")
                params[k] = v
        else:
            has_name_params = False
            base_provider_name = provider_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = provider_name
        else:
            # Determine the name we want this parameterized provider to have
            result_name = base_provider_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if (
            result_name in self.providers
            and not self.providers[result_name] is None
        ):
            p = self.providers[result_name]
            self.lock.release()
            return p

        with Project.ProviderLock(self, result_name):
            # Release the project lock, and continue with holding the part lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular provider name, no params (provider_name == result_name)
                if not provider_name in self.provider_configs:
                    # We don't know anything about such an provider
                    pc_logging.error(
                        "Provider '%s' not found in '%s'",
                        provider_name,
                        self.name,
                    )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_provider_config(provider_name)
                config = provider_config.ProviderConfiguration.normalize(
                    provider_name, config
                )
                factory.instantiate(
                    "provider", config["type"], self.ctx, self, self, config
                )

                if (
                    not provider_name in self.providers
                    or self.providers[provider_name] is None
                ):
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized provider %s"
                        % provider_name
                    )
                return self.providers[provider_name]

            # This provider has params (part_name != result_name)
            if not base_provider_name in self.providers:
                pc_logging.error(
                    "Base provider '%s' not found in '%s'",
                    base_provider_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base provider: %s" % base_provider_name)

            # Now we have the original assembly name and the complete set of parameters
            config = self.provider_configs[base_provider_name]
            if config is None:
                pc_logging.error(
                    "The config for the base provider '%s' is not found in '%s'",
                    base_provider_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_provider_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = provider_config.ProviderConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_provider_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
                    elif config["parameters"][param_name]["type"] == "array":
                        config["parameters"][param_name]["default"] = param_value
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the provider
            pc_logging.debug(
                "Initializing a parametrized provider: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized provider using the following config: %s"
            #     % pformat(config)
            # )
            factory.instantiate(
                "provider", config["type"], self.ctx, self, self, config
            )

            # See if it worked
            if not result_name in self.providers:
                pc_logging.error(
                    "Failed to instantiate parameterized assembly '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.providers[result_name]

    def get_suppliers(self, part_spec):
        # TODO(clirbee): return the eligible subset of suppliers
        # part_name = part_spec["name"]
        return self.suppliers

    def init_suppliers(self):
        cfg = self.config_obj.get("suppliers", {})
        if isinstance(cfg, str):
            cfg = {cfg: {}}
        elif isinstance(cfg, list):
            cfg = {c: {} for c in cfg}
        elif not isinstance(cfg, dict):
            pc_logging.error(
                "Invalid suppliers configuration in '%s': %s",
                self.name,
                str(cfg),
            )
            return

        self.suppliers = cfg

    def add_import(self, alias, location):
        if ":" in location:
            location_param = "url"
            if location.endswith(".tar.gz"):
                location_type = "tar"
            else:
                location_type = "git"
        else:
            location_param = "path"
            location_type = "local"

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            config = yaml.load(fp)
            fp.close()

        for elem in config:
            if elem == "import":
                imports = config["import"]
                imports[alias] = {
                    location_param: location,
                    "type": location_type,
                }
                break  # no need to iterate further
        with open(self.config_path) as fp:
            yaml.dump(config, fp)
            fp.close()

    def _validate_path(self, path, extension) -> tuple[bool, str, str]:
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        root = self.config_dir
        if not os.path.isabs(root):
            root = os.path.abspath(root)

        if not path.startswith(root):
            pc_logging.error("Can't add files outside of the package")
            return False, None, None

        path = path[len(root) :]
        if path[0] == os.path.sep:
            path = path[1:]

        name = path
        if name.endswith(".%s" % extension):
            name = name[: -len(extension) - 1]

        return True, path, name

    def _add_component(
        self,
        kind: str,
        path: str,
        section: str,
        ext_by_kind: dict[str, str],
        component_config,
    ) -> bool:
        if kind in ext_by_kind:
            ext = ext_by_kind[kind]
        else:
            ext = kind

        valid, path, name = self._validate_path(path, ext)
        if not valid:
            return False

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            config = yaml.load(fp)
            fp.close()

        obj = {"type": kind, **component_config}
        if name == path:
            obj["path"] = path

        found = False
        for elem in config:
            if elem == section:
                config_section = config[section]
                if config_section is None:
                    config_section = {}
                config_section[name] = obj
                config[section] = config_section
                found = True
                break  # no need to iterate further
        if not found:
            config[section] = {name: obj}

        with open(self.config_path, "w") as fp:
            yaml.dump(config, fp)
            fp.close()

        return True

    def add_part(self, kind: str, path: str, config={}) -> bool:
        pc_logging.info("Adding the part %s of type %s" % (path, kind))
        ext_by_kind = {
            "cadquery": "py",
            "build123d": "py",
            "ai-cadquery": "py",
            "ai-openscad": "scad",
        }
        return self._add_component(
            kind,
            path,
            "parts",
            ext_by_kind,
            config,
        )

    def add_assembly(self, kind: str, path: str, config={}) -> bool:
        pc_logging.info("Adding the assembly %s of type %s" % (path, kind))
        ext_by_kind = {}
        return self._add_component(
            kind,
            path,
            "assemblies",
            ext_by_kind,
            config,
        )

    def set_part_config(self, part_name, part_config):
        if "name" in part_config:
            del part_config["name"]
        if "orig_name" in part_config:
            del part_config["orig_name"]

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            package_config = yaml.load(fp)
            fp.close()

        if "parts" in package_config:
            parts = package_config["parts"]
            parts[part_name] = part_config
        else:
            package_config["parts"] = {part_name: part_config}

        with open(self.config_path, "w") as fp:
            yaml.dump(package_config, fp)
            fp.close()

    def update_part_config(
        self, part_name, part_config_update: dict[str, typing.Any]
    ):
        pc_logging.debug(
            "Updating part config: %s: %s" % (part_name, part_config_update)
        )
        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            config = yaml.load(fp)
            fp.close()

        if "parts" in config:
            parts = config["parts"]
            if part_name in parts:
                part_config = parts[part_name]
                for key, value in part_config_update.items():
                    if value is not None:
                        part_config[key] = value
                    else:
                        if key in part_config:
                            del part_config[key]

                with open(self.config_path, "w") as fp:
                    yaml.dump(config, fp)
                    fp.close()

    def test(self):
        for interface in self.interfaces.values():
            interface.test()
        for sketch in self.sketches.values():
            sketch.test()
        for part in self.parts.values():
            part.test()
        for assembly in self.assemblies.values():
            assembly.test()

    async def render_async(
        self,
        sketches=None,
        interfaces=None,
        parts=None,
        assemblies=None,
        format=None,
        output_dir=None,
    ):
        with pc_logging.Action("RenderPkg", self.name):
            # Override the default output_dir.
            # TODO(clairbee): pass the preference downstream without making a
            # persistent change.
            if not output_dir is None:
                if (
                    not "render" in self.config_obj
                    or self.config_obj["render"] is None
                ):
                    self.config_obj["render"] = {}
                self.config_obj["render"]["output_dir"] = output_dir

            render = self.config_obj.get("render", {})
            if render is None:
                render = {}

            # Enumerating all parts and assemblies
            if sketches is None:
                sketches = []
                if (
                    "sketches" in self.config_obj
                    and not self.config_obj["sketches"] is None
                ):
                    sketches = self.config_obj["sketches"].keys()
            if interfaces is None:
                interfaces = []
                if (
                    "interfaces" in self.config_obj
                    and not self.config_obj["interfaces"] is None
                ):
                    interfaces = self.config_obj["interfaces"].keys()
            if parts is None:
                parts = []
                if (
                    "parts" in self.config_obj
                    and not self.config_obj["parts"] is None
                ):
                    parts = self.config_obj["parts"].keys()
            if assemblies is None:
                assemblies = []
                if (
                    "assemblies" in self.config_obj
                    and not self.config_obj["assemblies"] is None
                ):
                    assemblies = self.config_obj["assemblies"].keys()

            # Enumerate
            shapes = []
            for name in sketches:
                shape = self.get_sketch(name)
                shapes.append(shape)
            for name in interfaces:
                shape = self.get_interface(name)
                # TODO(clairbee): interfaces are not yet renderable.
                # shapes.append(shape)
            for name in parts:
                shape = self.get_part(name)
                shapes.append(shape)
            for name in assemblies:
                shape = self.get_assembly(name)
                shapes.append(shape)

            # Render
            tasks = []
            for shape in shapes:
                shape_render = copy.copy(render)
                if (
                    "render" in shape.config
                    and not shape.config["render"] is None
                ):
                    shape_render = render_cfg_merge(
                        shape_render, shape.config["render"]
                    )

                # Determine which formats need to be rendered.
                # The format needs to be rendered either if it's mentioned in the config
                # or if it's explicitly requested in the params (e.g. comes from the
                # command line).
                if (
                    "svg" in shape_render
                    and shape_render["svg"] is not None
                    and not isinstance(shape_render["svg"], str)
                    and shape.kind
                    in shape_render.get("svg", {}).get("exclude", [])
                ):
                    render_svg = False
                elif format is None and "svg" in shape_render:
                    render_svg = True
                elif not format is None and format == "svg":
                    render_svg = True
                else:
                    render_svg = False

                if (
                    "png" in shape_render
                    and shape_render["png"] is not None
                    and not isinstance(shape_render["png"], str)
                    and shape.kind
                    in shape_render.get("png", {}).get("exclude", [])
                ):
                    render_png = False
                elif format is None and "png" in shape_render:
                    render_png = True
                elif not format is None and format == "png":
                    render_png = True
                else:
                    render_png = False

                if (
                    "step" in shape_render
                    and shape_render["step"] is not None
                    and not isinstance(shape_render["step"], str)
                    and shape.kind
                    in shape_render.get("step", {}).get("exclude", [])
                ):
                    render_step = False
                elif format is None and "step" in shape_render:
                    render_step = True
                elif not format is None and format == "step":
                    render_step = True
                else:
                    render_step = False

                if (
                    "stl" in shape_render
                    and shape_render["stl"] is not None
                    and not isinstance(shape_render["stl"], str)
                    and shape.kind
                    in shape_render.get("stl", {}).get("exclude", [])
                ):
                    render_stl = False
                elif format is None and "stl" in shape_render:
                    render_stl = True
                elif not format is None and format == "stl":
                    render_stl = True
                else:
                    render_stl = False

                if (
                    "3mf" in shape_render
                    and shape_render["3mf"] is not None
                    and not isinstance(shape_render["3mf"], str)
                    and shape.kind
                    in shape_render.get("3mf", {}).get("exclude", [])
                ):
                    render_3mf = False
                elif format is None and "3mf" in shape_render:
                    render_3mf = True
                elif not format is None and format == "3mf":
                    render_3mf = True
                else:
                    render_3mf = False

                if (
                    "threejs" in shape_render
                    and shape_render["threejs"] is not None
                    and not isinstance(shape_render["threejs"], str)
                    and shape.kind
                    in shape_render.get("threejs", {}).get("exclude", [])
                ):
                    render_threejs = False
                elif format is None and "threejs" in shape_render:
                    render_threejs = True
                elif not format is None and format == "threejs":
                    render_threejs = True
                else:
                    render_threejs = False

                if (
                    "obj" in shape_render
                    and shape_render["obj"] is not None
                    and not isinstance(shape_render["obj"], str)
                    and shape.kind
                    in shape_render.get("obj", {}).get("exclude", [])
                ):
                    render_obj = False
                elif format is None and "obj" in shape_render:
                    render_obj = True
                elif not format is None and format == "obj":
                    render_obj = True
                else:
                    render_obj = False

                if (
                    "gltf" in shape_render
                    and shape_render["gltf"] is not None
                    and not isinstance(shape_render["gltf"], str)
                    and shape.kind
                    in shape_render.get("gltf", {}).get("exclude", [])
                ):
                    render_gltf = False
                elif format is None and "gltf" in shape_render:
                    render_gltf = True
                elif not format is None and format == "gltf":
                    render_gltf = True
                else:
                    render_gltf = False

                if render_svg:
                    tasks.append(shape.render_svg_async(self.ctx, self))
                if render_png:
                    tasks.append(shape.render_png_async(self.ctx, self))
                if render_step:
                    tasks.append(shape.render_step_async(self.ctx, self))
                if render_stl:
                    tasks.append(shape.render_stl_async(self.ctx, self))
                if render_3mf:
                    tasks.append(shape.render_3mf_async(self.ctx, self))
                if render_threejs:
                    tasks.append(shape.render_threejs_async(self.ctx, self))
                if render_obj:
                    tasks.append(shape.render_obj_async(self.ctx, self))
                if render_gltf:
                    tasks.append(shape.render_gltf_async(self.ctx, self))

            await asyncio.gather(*tasks)

            if format == "readme" or (format is None and "readme" in render):
                self.render_readme_async(render, output_dir)

    def render(
        self,
        sketches=None,
        interfaces=None,
        parts=None,
        assemblies=None,
        format=None,
        output_dir=None,
    ):
        asyncio.run(
            self.render_async(
                sketches, interfaces, parts, assemblies, format, output_dir
            )
        )

    def render_readme_async(self, render_cfg, output_dir):
        if output_dir is None:
            output_dir = self.config_dir

        if render_cfg is None:
            render_cfg = {}
        cfg = render_cfg.get("readme", {})
        if isinstance(cfg, str):
            cfg = {"path": cfg}
        if cfg is None:
            cfg = {}

        path = os.path.join(output_dir, cfg.get("path", "README.md"))
        dir_path = os.path.dirname(path)
        return_path = os.path.relpath(output_dir, dir_path)

        exclude = cfg.get("exclude", [])
        if exclude is None:
            exclude = []

        name = self.name
        desc = self.desc
        docs = self.config_obj.get("docs", None)
        intro = None
        usage = None
        if docs:
            name = docs.get("name", name)
            intro = docs.get("intro", None)
            usage = docs.get("usage", None)

        lines = []
        lines += ["# %s" % name]
        lines += [""]
        if desc:
            lines += [desc]
            lines += [""]
        if intro:
            lines += [intro]
            lines += [""]

        if usage:
            lines += ["## Usage"]
            lines += [usage]
            lines += [""]

        if (
            self.config_obj.get("import", None) is not None
            and not "packages" in exclude
        ):
            imports = self.config_obj["import"]
            display_imports = []
            for alias in imports:
                if imports[alias].get("onlyInRoot", False):
                    continue
                display_imports.append(alias)

            if display_imports:
                lines += ["## Sub-Packages"]
                lines += [""]
                for alias in display_imports:
                    import_config = imports[alias]
                    columns = []

                    if (
                        "type" not in import_config
                        or import_config["type"] == "local"
                    ):
                        lines += [
                            "### [%s](./%s)"
                            % (
                                import_config["name"],
                                os.path.join(
                                    return_path,
                                    import_config["path"],
                                    "README.md",
                                ),
                            )
                        ]
                    elif import_config["type"] == "git":
                        lines += [
                            "### [%s](%s)"
                            % (import_config["name"], import_config["url"])
                        ]
                    else:
                        lines += ["### %s" % import_config["name"]]

                    if "desc" in import_config:
                        columns += [import_config["desc"]]
                    elif not columns:
                        columns += ["***Not documented yet.***"]

                    if len(columns) > 1:
                        lines += ["<table><tr>"]
                        lines += map(
                            lambda c: "<td valign=top>" + c + "</td>", columns
                        )
                        lines += ["</tr></table>"]
                    else:
                        lines += columns
                    lines += [""]

        def add_section(name, display_name, shape, render_cfg):
            config = shape.config

            if (
                "type" in config
                and config["type"] == "alias"
                and "aliases" in exclude
            ):
                return []

            path = None
            if "path" in config:
                path = config["path"]
            else:
                path = name
                if "type" in config:
                    if (
                        config["type"] == "cadquery"
                        or config["type"] == "build123d"
                        or config["type"] == "ai-cadquery"
                        or config["type"] == "ai-build123d"
                    ):
                        path += ".py"
                    elif (
                        config["type"] == "openscad"
                        or config["type"] == "ai-openscad"
                    ):
                        path += ".scad"
                    else:
                        path += "." + config["type"]

            columns = []
            if "svg" in render_cfg or (
                "type" in config and config["type"] == "svg"
            ):
                svg_cfg = render_cfg["svg"] if "svg" in render_cfg else {}
                if isinstance(svg_cfg, str):
                    svg_cfg = {"prefix": svg_cfg}
                svg_cfg = svg_cfg if svg_cfg is not None else {}
                image_path = os.path.join(
                    return_path,
                    svg_cfg.get("prefix", "."),
                    shape.name + ".svg",
                )
                test_image_path = os.path.join(
                    svg_cfg.get("prefix", "."),
                    shape.name + ".svg",
                )
                img_text = (
                    '<img src="%s" style="width: auto; height: auto; max-width: 200px; max-height: 200px;">'
                    % image_path
                )
                if path:
                    img_text = '<a href="%s">%s</a>' % (path, img_text)
                columns += [img_text]
            elif "png" in render_cfg:
                png_cfg = render_cfg["png"]
                png_cfg = png_cfg if png_cfg is not None else {}
                image_path = os.path.join(
                    return_path,
                    png_cfg.get("prefix", "."),
                    shape.name + ".png",
                )
                test_image_path = os.path.join(
                    png_cfg.get("prefix", "."),
                    shape.name + ".png",
                )
                img_text = (
                    '<img src="%s" style="width: auto; height: auto; max-width: 200px; max-height: 200px;">'
                    % image_path
                )
                if path:
                    img_text = '<a href="%s">%s</a>' % (path, img_text)
                columns += [img_text]
            else:
                image_path = None
                test_image_path = None

            if image_path is None or not os.path.exists(
                os.path.join(output_dir, test_image_path)
            ):
                pc_logging.warn(
                    "Skipping rendering of %s: no image found at %s"
                    % (name, test_image_path)
                )
                return []

            if "desc" in config:
                columns += [config["desc"]]

            if "parameters" in config:
                parameters = "Parameters:<br/><ul>\n"
                for param_name, param in config["parameters"].items():
                    if "enum" in param:
                        value = "<ul>\n"
                        for enum_value in param["enum"]:
                            if enum_value == param["default"]:
                                value += "<li><b>%s</b></li>\n" % enum_value
                            else:
                                value += "<li>%s</li>" % enum_value
                        value += "</ul>\n"
                    else:
                        value = param["default"]
                    parameters += "<li>%s: %s</li>\n" % (param_name, value)
                parameters += "</ul>\n"
                columns += [parameters]

            if "images" in config:
                images = "Input images:\n"
                for image in config["images"]:
                    images += (
                        '</br><img src="%s" alt="%s" style="width: auto; height: auto; max-width: 200px; max-height: 200px;" />\n'
                        % (
                            image,
                            image,
                        )
                    )
                columns += [images]

            if "aliases" in config:
                aliases = "Aliases:<br/><ul>"
                for alias in config["aliases"]:
                    aliases += "<li>%s</li>" % alias
                aliases += "</ul>"
                columns += [aliases]

            if hasattr(shape, "interfaces"):
                interfaces = "Interfaces:<br/>"
                for interface in shape.interfaces:
                    interfaces += "- %s<br/>" % interface.name
                columns += [interfaces]

            lines = ["### %s" % display_name]
            if len(columns) > 1:
                lines += ["<table><tr>"]
                lines += map(lambda c: "<td valign=top>" + c + "</td>", columns)
                lines += ["</tr></table>"]
            else:
                lines += columns
            lines += [""]
            return lines

        if self.assemblies and not "assemblies" in exclude:
            lines += ["## Assemblies"]
            lines += [""]
            shape_names = sorted(self.assemblies.keys())
            for name in shape_names:
                shape = self.assemblies[name]
                if shape.config["type"] == "alias":
                    source_path = normalize_resource_path(
                        self.name, shape.config["source_resolved"]
                    )
                    shape = self.ctx.get_assembly(source_path)
                    display_name = name + " (alias to " + shape.name + ")"
                else:
                    display_name = name
                lines += add_section(name, display_name, shape, render_cfg)

        if self.parts and not "parts" in exclude:
            lines += ["## Parts"]
            lines += [""]
            shape_names = sorted(self.parts.keys())
            for name in shape_names:
                shape = self.parts[name]
                if shape.config["type"] == "alias":
                    source_path = normalize_resource_path(
                        self.name, shape.config["source_resolved"]
                    )
                    shape = self.ctx.get_part(source_path)
                    display_name = name + " (alias to " + shape.name + ")"
                else:
                    display_name = name
                lines += add_section(name, display_name, shape, render_cfg)

        if self.interfaces and not "interfaces" in exclude:
            lines += ["## Interfaces"]
            lines += [""]
            shape_names = sorted(self.interfaces.keys())
            for name in shape_names:
                shape = self.interfaces[name]
                lines += add_section(name, name, shape, render_cfg)

        if self.sketches and not "sketches" in exclude:
            lines += ["## Sketches"]
            lines += [""]
            shape_names = sorted(self.sketches.keys())
            for name in shape_names:
                shape = self.sketches[name]
                lines += add_section(name, name, shape, render_cfg)

        lines += [
            "<br/><br/>",
            "",
            "*Generated by [PartCAD](https://partcad.org/)*",
        ]

        lines = map(lambda s: s + "\n", lines)

        f = open(path, "w")
        f.writelines(lines)
        f.close()



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from . import factory
from .provider import Provider


class ProviderFactory(factory.Factory):
    provider: Provider
    name: str
    orig_name: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config: object,
    ):
        super().__init__()

        self.ctx = ctx
        self.project = source_project
        self.config = config

        self.target_project = target_project
        self.name = config["name"]
        self.orig_name = config["orig_name"]

    def _create_provider(self, config: object) -> Provider:
        provider = Provider(config)
        provider.name = f"{self.target_project.name}:{self.name}"
        provider.config = self.config
        provider.project_name = (
            self.target_project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for provider_factory_file only
        provider.info = lambda: self.info(provider)
        return provider

    def _create(self, config: object):
        self.provider = self._create_provider(config)
        if hasattr(self, "path"):
            self.provider.path = self.path

        self.target_project.providers[self.name] = self.provider
        self.ctx.stats_providers += 1



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import base64
import importlib
import mimetypes
from pathlib import Path
import re
import threading
from typing import Any

# Lazy-load AI imports as they are not always needed
# import openai as openai_genai
openai_genai = None

from . import logging as pc_logging
from .user_config import user_config

lock = threading.Lock()
OPENAI_API_KEY = None
openai_client = None

model_tokens = {
    "gpt-3": 4000,
    "gpt-3.5-turbo": 4096,
    "gpt-4": 8000,  # 32600,
    "gpt-4-vision-preview": 8192,
    "gpt-4o": 16000,  # 32600,
    "gpt-4o-mini": 16000,  # 32600,
}


def openai_once():
    global OPENAI_API_KEY, openai_client, openai_genai

    with lock:
        if openai_genai is None:
            openai_genai = importlib.import_module("openai")

        latest_key = user_config.openai_api_key
        if latest_key != OPENAI_API_KEY:
            OPENAI_API_KEY = latest_key
            if not OPENAI_API_KEY is None:
                openai_genai.api_key = OPENAI_API_KEY
                openai_client = openai_genai.OpenAI(api_key=OPENAI_API_KEY)
                return True

        if OPENAI_API_KEY is None:
            raise Exception("OpenAI API key is not set")

    return True


class AiOpenAI:
    def generate_openai(
        self,
        model: str,
        prompt: str,
        config: dict[str, Any] = {},
        options_num: int = 1,
    ):
        if not openai_once():
            return None

        if "tokens" in config:
            tokens = config["tokens"]
        else:
            tokens = model_tokens[model]

        if "top_p" in config:
            top_p = config["top_p"]
        else:
            top_p = None

        if "temperature" in config:
            temperature = config["temperature"]
        else:
            temperature = None

        pc_logging.debug("Prompt: %s", prompt)

        image_content = []

        def insert_image(match):
            filename = match.group(1)
            image_content.append(
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "data:%s;base64,%s"
                        % (
                            mimetypes.guess_type(filename, False)[0],
                            base64.b64encode(
                                Path(filename).read_bytes()
                            ).decode(),
                        ),
                        "detail": "high",
                    },
                }
            )
            return "IMAGE_INSERTED_HERE"

        prompt = re.sub(r"INSERT_IMAGE_HERE\(([^)]*)\)", insert_image, prompt)
        text_content = list(
            map(
                lambda prompt_section: {"type": "text", "text": prompt_section},
                prompt.split("IMAGE_INSERTED_HERE"),
            )
        )

        content = []
        for i in range(len(text_content)):
            content.append(text_content[i])
            if i < len(image_content):
                content.append(image_content[i])

        cc = openai_client.chat.completions.create(
            messages=[
                {"role": "user", "content": content},
            ],
            stream=False,
            n=options_num,
            max_tokens=tokens,
            top_p=top_p,
            temperature=temperature,
            model=model,
        )

        products = []
        try:
            for choice in cc.choices:
                if hasattr(choice, "role") and choice.role == "system":
                    continue

                script = choice.message.content

                products.append(script)
        except Exception as e:
            pc_logging.exception(e)

        return products



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-21
#
# Licensed under Apache License, Version 2.0.
#

from .geom import Location
from . import logging as pc_logging
from .utils import resolve_resource_path


class InterfaceInherits:
    """One of the other interface types inherited by this interface.
    May include multiple instances of each inherited interface.

    This object is contained in the 'inherits' section of an interface
    or the 'implements' section of a part/assembly.
    It points at the interface object that already exists in the package."""

    source_project_name: str
    source_interface_name: str
    interface = None
    instances: dict[str, Location]

    def __init__(self, name, project, config: dict = {}):
        self.instances = {}
        # Resolve the interface by name
        if ":" in name:
            self.source_project_name, self.source_interface_name = (
                resolve_resource_path(
                    project.name,
                    name,
                )
            )
        else:
            self.source_project_name = project.name
            self.source_interface_name = name

        self.name = self.source_project_name + ":" + self.source_interface_name
        pc_logging.debug("Fetching the interface %s" % self.name)

        if self.source_project_name == project.name:
            target_project = project
        else:
            target_project = project.ctx.get_project(self.source_project_name)
            if target_project is None:
                raise Exception(
                    "Failed to find the project to inherit the interface from: %s"
                    % self.source_project_name
                )
        self.interface = target_project.get_interface(
            self.source_interface_name
        )

        if self.interface is None:
            raise Exception(
                "Failed to find the interface to inherit: %s" % name
            )

        pc_logging.debug("Inherited config: %s: %s" % (name, str(config)))
        if config is None:
            config = {"": None}
        elif isinstance(config, str):
            config = {config: None}
        elif isinstance(config, list):
            if (
                len(config) == 3
                and isinstance(config[0], list)
                and len(config[0]) == 3
                and isinstance(config[1], list)
                and len(config[1]) == 3
                and (isinstance(config[2], int) or isinstance(config[2], float))
            ):
                # This is a single instance with a location
                config = {"": config}
            else:
                # This is a list of instances without locations
                # TODO(clairbee): is this practical at all?
                config = {instance_name: None for instance_name in config}
        elif not isinstance(config, dict):
            raise Exception(
                "Invalid 'inherits' section in the interface '%s'" % self.name
            )
        pc_logging.debug(
            "Normalized inherited config: %s: %s" % (name, str(config))
        )

        for instance_name, instance_config in config.items():
            offsets = []

            if isinstance(instance_config, dict):
                instance_location_config = instance_config.get("location", None)

                params = instance_config.get("params", {})
                for param_name, param_value in params.items():
                    param = self.interface.params.get(param_name, None)
                    if param is not None:
                        # Extend the list of offsets to apply, if any
                        offsets.extend(param.get_offsets(param_value))
            else:
                instance_location_config = instance_config

            if instance_location_config is None:
                location = Location()
            else:
                location = Location(instance_location_config)

            # TODO(clairbee): consider resolving the parameters here: apply the offsets and drop the parameters
            # Do not apply the offsets yet?
            # if len(offsets) > 0:
            #     trsf = location.wrapped.Transformation()
            #     for offset in offsets:
            #         trsf.Multiply(offset)
            #         # trsf.Multiply(offset.Inverted())
            #     location = Location(trsf)

            self.instances[instance_name] = location
